import os
import math
import numpy as np
import pandas as pd
from tqdm import tqdm
from time import time
from pathlib import Path

import torch
import torch.nn as nn

from ..utils import data_utils as du
from ..utils import training_utils as tu
from ..utils import logging_utils as lu


def find_idx(array, value):
    """Utility to find the index of the element of ``array`` that most closely
    matches ``value``

    Args:
        array (np.array): The array in which to search
        value (float): The value for which we are looking for a match

    Returns:
        (int) the index of of the element of ``array`` that most closely
        matches ``value``

    """

    idx = np.searchsorted(array, value, side="left")
    if idx > 0 and (
        idx == len(array)
        or math.fabs(value - array[idx - 1]) < math.fabs(value - array[idx])
    ):
        return idx - 1
    else:
        return idx


def get_batch_predictions(rnn, X, target):
    """Utility to obtain predictions for a given batch

    Args:
        rnn (torch.nn): The RNN model
        X (torch.Tensor): The batch on which to carry out predictions
        target (torch.longTensor): The true class of each element in the batch

    Returns:
        Tuple containing

        - arr_preds (np.array): predictions
        - arr_target (np.array): actual targets

    """

    out = rnn.forward(X)
    arr_preds = nn.functional.softmax(out, dim=-1).data.cpu().numpy()
    arr_target = target.detach().cpu().numpy()

    return arr_preds, arr_target


def get_batch_predictions_MFE(rnn, X, target):
    """Utility to obtain predictions for a given batch

    Args:
        rnn (torch.nn): The RNN model
        X (torch.Tensor): The batch on which to carry out predictions
        target (torch.longTensor): The true class of each element in the batch

    Returns:
        Tuple containing

        - arr_preds (np.array): predictions
        - arr_target (np.array): actual targets

    """

    out = rnn.forward(X, mean_field_inference=True)
    arr_preds = nn.functional.softmax(out, dim=-1).data.cpu().numpy()
    arr_target = target.detach().cpu().numpy()

    return arr_preds, arr_target


def get_predictions(settings, model_file=None):
    """Obtain predictions for a given RNN model specified by the
    ``settings`` argument or alternatively, by a model_file

    - Models are benchmarked on the test data set
    - Batch size can be controled to speed up predictions
    - For Bayesian models, multiple predictions are carried to
        obtain a distribution of predictions
    - Predictions are computed for full lightcurves, and around the peak light
    - Predictions are saved to a pickle file (for faster loading)

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters
        model_file (str): Path to saved model weights. Default: ``None``
    """

    settings.random_length = False
    settings.random_redshift = False

    if "vanilla" in settings.pytorch_model_name:
        settings.num_inference_samples = 1

    # Load RNN model
    rnn = tu.get_model(settings, len(settings.training_features))
    if model_file is None:
        dump_dir = f"{settings.models_dir}/{settings.pytorch_model_name}"
        model_file = f"{dump_dir}/{settings.pytorch_model_name}.pt"
    else:
        dump_dir = f"{settings.dump_dir}/models/{settings.pytorch_model_name}"
        os.makedirs(dump_dir, exist_ok=True)

    if settings.override_source_data is not None:
        settings.source_data = settings.override_source_data
        settings.set_pytorch_model_name()

    prediction_file = (
        f"{dump_dir}/PRED_{settings.pytorch_model_name}.pickle"
    )

    rnn_state = torch.load(model_file, map_location=lambda storage, loc: storage)
    rnn.load_state_dict(rnn_state)
    rnn.to(settings.device)
    rnn.eval()

    # Load the data
    list_data_test = tu.load_HDF5(settings, test=True)
    
    # Batching stuff together
    num_elem = len(list_data_test)
    factor = (num_elem // 4) if settings.source_data == "photometry" else num_elem
    num_batches = num_elem / min(factor,100000)
    list_batches = np.array_split(np.arange(num_elem), num_batches)
    
    # Prepare output arrays
    d_pred = {
        key: np.zeros(
            (num_elem, settings.num_inference_samples, settings.nb_classes)
        ).astype(np.float32)
        for key in [
            "all",
            "PEAKMJD-2",
            "PEAKMJD-1",
            "PEAKMJD",
            "PEAKMJD+1",
            "PEAKMJD+2",
        ]
        + [f"all_{OOD}" for OOD in du.OOD_TYPES]
    }
    d_pred["target"] = np.zeros((num_elem, settings.num_inference_samples)).astype(
        np.int64
    )
    d_pred["SNID"] = np.zeros((num_elem, settings.num_inference_samples)).astype(
        np.str
    )

    d_pred_MFE = {
        key: np.zeros((num_elem, 1, settings.nb_classes)).astype(np.float32)
        for key in ["all"] + [f"all_{OOD}" for OOD in du.OOD_TYPES]
    }
    d_pred_MFE["target"] = np.zeros((num_elem, 1)).astype(np.int64)
    d_pred_MFE["SNID"] = np.zeros((num_elem, 1)).astype(np.str)

    # Fetch SN info
    df_SNinfo = du.load_HDF5_SNinfo(settings).set_index("SNID")

    # Loop over data and make prediction
    for batch_idxs in tqdm(
        list_batches, desc="Computing predictions on test set", ncols=100
    ):

        start_idx, end_idx = batch_idxs[0], batch_idxs[-1] + 1
        SNIDs = [data[2] for data in list_data_test[start_idx:end_idx]]

        peak_MJDs = df_SNinfo.loc[SNIDs]["PEAKMJDNORM"].values
        delta_times = [
            data[3][:, settings.d_feat_to_idx["delta_time"]]
            for data in list_data_test[start_idx:end_idx]
        ]
        times = [np.cumsum(t) for t in delta_times]

        with torch.no_grad():

            #############################
            # Full lightcurve prediction
            #############################

            packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
                list_data_test, batch_idxs, settings
            )

            for iter_ in tqdm(range(settings.num_inference_samples), ncols=100):

                arr_preds, arr_target = get_batch_predictions(
                    rnn, packed, target_tensor
                )

                # Rever sorting that occurs in get_batch_predictions
                arr_preds = arr_preds[idxs_rev_sort]
                arr_target = arr_target[idxs_rev_sort]

                d_pred["all"][start_idx:end_idx, iter_] = arr_preds
                d_pred["target"][start_idx:end_idx, iter_] = arr_target
                d_pred["SNID"][start_idx:end_idx, iter_] = SNIDs

            # MFE
            arr_preds, arr_target = get_batch_predictions_MFE(
                rnn, packed, target_tensor
            )

            # Rever sorting that occurs in get_batch_predictions
            arr_preds = arr_preds[idxs_rev_sort]
            arr_target = arr_target[idxs_rev_sort]

            d_pred_MFE["all"][start_idx:end_idx, 0] = arr_preds
            d_pred_MFE["target"][start_idx:end_idx, 0] = arr_target
            d_pred_MFE["SNID"][start_idx:end_idx, 0] = SNIDs

            #############################
            # Predictions around PEAKMJD
            #############################
            for offset in [-2, -1, 0, 1, 2]:
                slice_idxs = [
                    find_idx(times[k], peak_MJDs[k] + offset) for k in range(len(times))
                ]
                # Split in 2 arrays:
                # oob_idxs: the slice for early prediction is empty for those indices
                # inb_idxs: the slice is not empty
                oob_idxs = np.where(np.array(slice_idxs) < 1)[0]
                inb_idxs = np.where(np.array(slice_idxs) >= 1)[0]

                if len(inb_idxs)>0:
                    # We only carry out prediction for samples in ``inb_idxs``
                    offset_batch_idxs = [batch_idxs[b] for b in inb_idxs]
                    max_lengths = [slice_idxs[b] for b in inb_idxs]
                    packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
                        list_data_test, offset_batch_idxs, settings, max_lengths=max_lengths
                    )

                    for iter_ in tqdm(range(settings.num_inference_samples), ncols=100):

                        arr_preds, arr_target = get_batch_predictions(
                            rnn, packed, target_tensor
                        )

                        # Rever sorting that occurs in get_batch_predictions
                        arr_preds = arr_preds[idxs_rev_sort]

                        suffix = str(offset) if offset != 0 else ""
                        suffix = f"+{suffix}" if offset > 0 else suffix
                        col = f"PEAKMJD{suffix}"

                        d_pred[col][start_idx + inb_idxs, iter_] = arr_preds
                        # For oob_idxs, no prediction can be made, fill with nan
                        d_pred[col][start_idx + oob_idxs, iter_] = np.nan

            #############################
            # OOD predictions
            #############################

            for OOD in ["random", "shuffle", "reverse", "sin"]:
                packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
                    list_data_test, batch_idxs, settings, OOD=OOD
                )

                for iter_ in tqdm(range(settings.num_inference_samples), ncols=100):

                    arr_preds, arr_target = get_batch_predictions(
                        rnn, packed, target_tensor
                    )

                    # Revert sorting that occurs in get_batch_predictions
                    arr_preds = arr_preds[idxs_rev_sort]
                    arr_target = arr_target[idxs_rev_sort]

                    d_pred[f"all_{OOD}"][start_idx:end_idx, iter_] = arr_preds

                arr_preds, arr_target = get_batch_predictions_MFE(
                    rnn, packed, target_tensor
                )

                # Revert sorting that occurs in get_batch_predictions
                arr_preds = arr_preds[idxs_rev_sort]
                arr_target = arr_target[idxs_rev_sort]

                d_pred_MFE[f"all_{OOD}"][start_idx:end_idx, 0] = arr_preds

    # Flatten all arrays and aggregate in dataframe
    d_series = {}
    for (key, value) in d_pred.items():
        value = value.reshape((num_elem * settings.num_inference_samples, -1))
        value_dim = value.shape[1]
        if value_dim == 1:
            d_series[key] = np.ravel(value)
        else:
            for i in range(value_dim):
                d_series[f"{key}_class{i}"] = value[:, i]
    df_pred = pd.DataFrame.from_dict(d_series)

    # Flatten all arrays and aggregate in dataframe
    d_series_MFE = {}
    for (key, value) in d_pred_MFE.items():
        value = value.reshape((num_elem * 1, -1))
        value_dim = value.shape[1]
        if value_dim == 1:
            d_series_MFE[key] = np.ravel(value)
        else:
            for i in range(value_dim):
                d_series_MFE[f"{key}_class{i}"] = value[:, i]
    df_pred_MFE = pd.DataFrame.from_dict(d_series_MFE)

    # Save predictions
    df_pred.to_pickle(prediction_file)

    # Saving aggregated preds for bayesian models
    if settings.model == 'variational' or settings.model == 'bayesian':
        med_pred = df_pred.groupby("SNID").median()
        med_pred.columns = [str(col) + '_median' for col in med_pred.columns]
        std_pred = df_pred.groupby("SNID").std()
        std_pred.columns = [str(col) + '_std' for col in std_pred.columns]
        df_bayes = pd.merge(med_pred,std_pred,on="SNID")
        df_bayes["SNID"] = df_bayes.index
        df_bayes["target"] = df_bayes["target_median"]
        bay_pred_file = prediction_file.replace(".pickle","_aggregated.pickle")
        df_bayes.to_pickle(bay_pred_file)

    g_pred = df_pred.groupby("SNID").median()
    preds = g_pred[[f"all_class{i}" for i in range(settings.nb_classes)]].values
    preds = np.argmax(preds, 1)
    acc = (preds == g_pred.target.values).sum() / len(g_pred)

    # Display accuracy
    lu.print_green("Full Accuracy", acc)
    for col in [f"PEAKMJD{s}" for s in du.OFFSETS_STR]:

        preds_target = g_pred[
            [f"{col}_class{i}" for i in range(settings.nb_classes)] + ["target"]
        ].dropna()
        preds = preds_target[
            [f"{col}_class{i}" for i in range(settings.nb_classes)]
        ].values
        target = preds_target["target"].values
        preds = np.argmax(preds, 1)
        acc = (preds == target).sum() / len(g_pred)

        lu.print_green(f"{col} Accuracy", acc)

    print()
    print()

    class_col = [f"all_class{i}" for i in range(settings.nb_classes)]
    tmp = df_pred[["SNID", "target"] + class_col].groupby("SNID").mean()
    preds = np.argmax(tmp[class_col].values, 1)
    acc = (preds == tmp.target.values).sum() / len(tmp)
    lu.print_green(f"Accuracy MC", acc)

    for OOD in ["random", "reverse", "shuffle", "sin"]:
        class_col_ood = [f"all_{OOD}_class{i}" for i in range(settings.nb_classes)]
        entropy_ood = (
            -(df_pred[class_col_ood].values * np.log(df_pred[class_col_ood].values))
            .sum(1)
            .mean()
        )
        entropy = (
            -(df_pred[class_col].values * np.log(df_pred[class_col].values))
            .sum(1)
            .mean()
        )
        lu.print_green(f"Delta Entropy {OOD} MC", entropy_ood - entropy)

    print()
    print()

    tmp = df_pred_MFE[["SNID", "target"] + class_col].groupby("SNID").mean()
    preds = np.argmax(tmp[class_col].values, 1)
    acc = (preds == tmp.target.values).sum() / len(tmp)
    lu.print_green(f"Accuracy MFE", acc)

    for OOD in ["random", "reverse", "shuffle", "sin"]:
        class_col_ood = [f"all_{OOD}_class{i}" for i in range(settings.nb_classes)]
        entropy_ood = (
            -(
                df_pred_MFE[class_col_ood].values
                * np.log(df_pred_MFE[class_col_ood].values)
            )
            .sum(1)
            .mean()
        )
        entropy = (
            -(df_pred_MFE[class_col].values * np.log(df_pred_MFE[class_col].values))
            .sum(1)
            .mean()
        )
        lu.print_green(f"Delta Entropy {OOD} MFE", entropy_ood - entropy)

    lu.print_green("Finished getting predictions ")

    return prediction_file


def get_predictions_for_speed_benchmark(settings):
    """Test RNN models inference speed

    - Models are benchmarked on the test data set
    - Batch size can be controled to speed up predictions
    - For Bayesian models, multiple predictions are carried to
        obtain a distribution of predictions
    - Results are saved to a .csv for future use

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters
    """

    settings.random_length = False

    if "vanilla" in settings.pytorch_model_name:
        settings.num_inference_samples = 1

    # Load the data
    list_data_test = tu.load_HDF5(settings, test=True)

    # Load RNN model
    rnn = tu.get_model(settings, len(settings.training_features))
    rnn.to(settings.device)
    rnn.eval()

    # Batching lightcurves
    num_elem = len(list_data_test)
    num_batches = num_elem / min(num_elem, settings.batch_size)
    list_batches = np.array_split(np.arange(num_elem), num_batches)

    start_time = time()

    # Loop over data and predict
    for batch_idxs in tqdm(list_batches, ncols=100):

        with torch.no_grad():

            #############################
            # Full lightcurve prediction
            #############################

            packed, _, target_tensor, _ = tu.get_data_batch(
                list_data_test, batch_idxs, settings
            )

            for iter_ in tqdm(range(settings.num_inference_samples), ncols=100):

                arr_preds, arr_target = get_batch_predictions(
                    rnn, packed, target_tensor
                )

    total_time = time() - start_time
    supernova_per_s = num_elem / total_time
    model_id = f"{settings.model}_{settings.batch_size}_{settings.device}"

    df = pd.DataFrame(
        data=np.array([supernova_per_s]).astype(np.float16), columns=["Supernova_per_s"]
    )
    df["model"] = settings.model
    df["batch_size"] = settings.batch_size
    df["device"] = settings.device
    df["id"] = model_id

    # Save results to csv
    results_file = Path(settings.stats_dir) / "rnn_speed.csv"
    try:
        df_all = pd.read_csv(results_file, index_col="id")
        df_all.loc[model_id] = df.set_index("id").loc[model_id]
        df_all.reset_index().to_csv(results_file, index=False)
    except Exception:
        df.to_csv(results_file, index=False)



import os
import h5py
import glob
import shlex
import subprocess
import numpy as np
import pandas as pd
from tqdm import tqdm
import multiprocessing
from pathlib import Path
from functools import partial

from ..utils import data_utils
from ..utils import logging_utils

from concurrent.futures import ProcessPoolExecutor


def preprocess_plasticc(settings, df):
    """

    """

    # Rename columns
    df = df.rename(
        columns={
            "object_id": "SNID",
            "passband": "FLT",
            "flux": "FLUXCAL",
            "flux_err": "FLUXCALERR",
            "mjd": "MJD",
            "target": "raw_target",
            "hostgal_photoz": "HOSTGAL_PHOTOZ",
            "hostgal_photoz_err": "HOSTGAL_PHOTOZ_ERR",
            "hostgal_specz": "HOSTGAL_SPECZ",
        }
    )
    df.FLT = df.FLT.replace(data_utils.DICT_PLASTICC_FILTERS)

    df["HOSTGAL_PHOTOZ"] = df["HOSTGAL_PHOTOZ"].fillna(-1)
    df["HOSTGAL_PHOTOZ_ERR"] = df["HOSTGAL_PHOTOZ_ERR"].fillna(-1)
    df["HOSTGAL_SPECZ"] = df["HOSTGAL_SPECZ"].fillna(-1)

    df = data_utils.compute_delta_time(df)
    df["rel_time"] = df[["SNID", "delta_time"]].groupby("SNID").cumsum()

    if "raw_target" in df.columns.values.tolist():
        df["target"] = df.raw_target.replace(data_utils.DICT_PLASTICC_CLASS)

    return df


def pivot_dataframe(df):
    """
    """

    list_filters = data_utils.PLASTICC_FILTERS

    arr_MJD = df.MJD.values
    arr_delta_time = df.delta_time.values

    # Loop over times to create grouped MJD:
    # if filters are acquired within less than 0.33 MJD (~8 hours) of each other
    # they get assigned the same time
    min_dt = 0.33
    time_last_change = 0
    arr_grouped_MJD = np.zeros_like(arr_MJD)
    for i in range(len(df)):
        time = arr_MJD[i]
        dt = arr_delta_time[i]
        # 2 possibilities to update the time
        # dt == 0 (it"s a new light curve)
        # time - time_last_change > min_dt
        if dt == 0 or (time - time_last_change) > min_dt:
            arr_grouped_MJD[i] = time
            time_last_change = time
        else:
            arr_grouped_MJD[i] = arr_grouped_MJD[i - 1]

    # Add grouped delta time to dataframe
    df["grouped_MJD"] = np.array(arr_grouped_MJD)

    # Some filters may appear multiple times with the same grouped MJD within same light curve
    # When this happens, we select the one with lowest FLUXCALERR
    df = df.sort_values("FLUXCALERR").groupby(["SNID", "grouped_MJD", "FLT"]).first()
    # We then reset the index
    df = df.reset_index()

    # drop columns that won"t be used onwards
    df = df.drop(
        [
            "MJD",
            "delta_time",
            "rel_time",
            "ra",
            "decl",
            "gal_b",
            "gal_l",
            "detected",
            "mwebv",
            "ddf",
            "distmod",
        ],
        1,
    )

    group_features_list = ["SNID", "grouped_MJD"]
    redshift_features = [k for k in df.keys() if "HOST" in k]

    if "target" in df.columns.values.tolist():
        group_features_list += ["target", "raw_target"]

    # Create a df of redshift information.
    # This is because sometimes, redshift is not available (NaN)
    # NaN create subtle bugs when pivoting !
    # We'll add the redshift info back to the pivoted dataframe afterwards
    # For a given lightcurve, redshift is fixed, so we can groupby and apply first
    # to create a dataframe containing one line per SNID, with the redshift
    df_redshift = df[["SNID"] + redshift_features].groupby("SNID").first()

    df = df.drop(redshift_features, axis=1)

    # Pivot so that for a given MJD, we have info on all available fluxes / error
    SNID_before_pivot = set(df.SNID.unique().tolist())
    df = pd.pivot_table(df, index=group_features_list, columns=["FLT"])

    # Flatten columns
    df.columns = ["_".join(col).strip() for col in df.columns.values]
    # Reset index to get grouped_MJD and target as columns
    cols_to_reset_list = [c for c in df.index.names if c != "SNID"]
    df.reset_index(cols_to_reset_list, inplace=True)
    # Rename grouped_MJD to MJD
    df.rename(columns={"grouped_MJD": "MJD"}, inplace=True)

    SNID_after_pivot = set(np.unique(df.index.values).tolist())
    assert SNID_before_pivot == SNID_after_pivot

    df = df.join(df_redshift, how="left")

    # New column to indicate which filter is present
    # The column will read ``rg`` if r,g are present; ``rgz`` if r,g and z are present, etc.
    for flt in list_filters:
        df[flt] = np.where(df[f"FLUXCAL_{flt}"].isnull(), "", flt)
    df["FLT"] = df[list_filters[0]]
    for flt in list_filters[1:]:
        df["FLT"] += df[flt]
    # Drop some irrelevant columns
    df = df.drop(list_filters, 1)
    # Finally replace NaN with 0
    df = df.fillna(0)
    # Add delta_time back. We removed all delta time columns above as they get
    # filled with NaN during pivot. It is clearer to recompute delta time once the pivot is complete
    df = data_utils.compute_delta_time(df)

    # Cast columns to float32 to save space
    for c in df.columns:
        if df[c].dtype == np.float64:
            df[c] = df[c].astype(np.float32)

    df.drop(columns="MJD", inplace=True)

    return df


def save_to_HDF5(settings, df, hdf5_file_name=None, verbose=True):
    """
    """
    list_training_features = settings.training_features_to_normalize + [
        "HOSTGAL_PHOTOZ",
        "HOSTGAL_PHOTOZ_ERR",
        "HOSTGAL_SPECZ",
    ]

    list_training_features
    assert df.index.name == "SNID", "SNID must be set as index"

    # Get the list of lightcurve IDs
    ID = df.index.values
    # Find out when ID changes => find start and end idx of each lightcurve
    idx_change = np.where(ID[1:] != ID[:-1])[0] + 1
    idx_change = np.hstack(([0], idx_change, [len(df)]))
    list_start_end = [(s, e) for s, e in zip(idx_change[:-1], idx_change[1:])]
    # N.B. We could use df.loc[SNID], more elegant but much slower

    logging_utils.print_green(
        f"Found {len(np.unique(ID))} unique lightcurves in dataset", verbose=verbose
    )

    # # Shuffle
    # np.random.shuffle(list_start_end)

    # Save hdf5 file
    if hdf5_file_name is None:
        hdf5_file_name = settings.hdf5_file_name
    with h5py.File(hdf5_file_name, "w") as hf:

        n_samples = len(list_start_end)

        # These arrays can be filled in one shot
        start_idxs = [i[0] for i in list_start_end]
        shuffled_ID = ID[start_idxs]
        hf.create_dataset(
            "SNID", data=shuffled_ID.astype(np.int32), dtype=np.dtype("int32")
        )

        logging_utils.print_green("Saving class", verbose=verbose)
        # only save target, splits, normalizations when building the training database
        if "target" in df.columns.values.tolist():

            # Save target info
            class_columns = ["target", "raw_target"]
            for class_column in class_columns:
                hf.create_dataset(
                    class_column,
                    data=df[class_column].values[start_idxs],
                    dtype=np.dtype("int64"),
                )
                # 0 if trainig set, 1 if valid set
                split_idxs = np.random.binomial(
                    1, 0.1, size=(len(list_start_end),)
                ).astype(np.int64)
                df = df.drop(class_column, axis=1)

            # Save dataset split
            logging_utils.print_green("Saving splits", verbose=verbose)
            hf.create_dataset("dataset", data=split_idxs, dtype=np.dtype("int64"))

            ########################
            # Normalize per feature
            ########################
            logging_utils.print_green("Compute normalizations", verbose=verbose)
            gnorm = hf.create_group("normalizations")

            # using normalization per feature
            for feat in settings.training_features_to_normalize:

                # Log transform plus mean subtraction and standard dev subtraction
                log_standardized = data_utils.log_standardization(df[feat].values)
                # Store normalization parameters
                gnorm.create_dataset(f"{feat}/min", data=log_standardized.arr_min)
                gnorm.create_dataset(f"{feat}/mean", data=log_standardized.arr_mean)
                gnorm.create_dataset(f"{feat}/std", data=log_standardized.arr_std)

            #####################################
            # Normalize flux and fluxerr globally
            #####################################
            logging_utils.print_green("Compute global normalizations", verbose=verbose)
            gnorm = hf.create_group("normalizations_global")

            ################
            # FLUX features
            #################
            flux_features = [f"FLUXCAL_{f}" for f in data_utils.PLASTICC_FILTERS]
            flux_log_standardized = data_utils.log_standardization(
                df[flux_features].values
            )
            # Store normalization parameters
            gnorm.create_dataset(f"FLUXCAL/min", data=flux_log_standardized.arr_min)
            gnorm.create_dataset(f"FLUXCAL/mean", data=flux_log_standardized.arr_mean)
            gnorm.create_dataset(f"FLUXCAL/std", data=flux_log_standardized.arr_std)

            ###################
            # FLUXERR features
            ###################
            fluxerr_features = [f"FLUXCALERR_{f}" for f in data_utils.PLASTICC_FILTERS]
            fluxerr_log_standardized = data_utils.log_standardization(
                df[fluxerr_features].values
            )
            # Store normalization parameters
            gnorm.create_dataset(
                f"FLUXCALERR/min", data=fluxerr_log_standardized.arr_min
            )
            gnorm.create_dataset(
                f"FLUXCALERR/mean", data=fluxerr_log_standardized.arr_mean
            )
            gnorm.create_dataset(
                f"FLUXCALERR/std", data=fluxerr_log_standardized.arr_std
            )

        ####################################
        # Save the rest of the data to hdf5
        ####################################

        # This type allows one to store flat arrays of variable
        # length inside an HDF5 group
        data_type = h5py.special_dtype(vlen=np.dtype("float32"))

        hf.create_dataset("data", (n_samples,), dtype=data_type)

        # Remove FLT column
        df = df.drop("FLT", axis=1)

        # store feature names
        list_training_features = df.columns.values.tolist()
        hf.create_dataset(
            "features",
            (len(list_training_features),),
            dtype=h5py.special_dtype(vlen=str),
        )
        hf["features"][:] = list_training_features
        logging_utils.print_green(
            "Saved features:", ",".join(list_training_features), verbose=verbose
        )

        # Save training features to hdf5
        logging_utils.print_green("Save data features to HDF5", verbose=verbose)
        arr_feat = df[list_training_features].values
        hf["data"].attrs["n_features"] = len(list_training_features)
        iterator = (
            enumerate(tqdm(list_start_end, desc="Filling hdf5", ncols=100))
            if verbose
            else enumerate(list_start_end)
        )
        for idx, idx_pair in iterator:
            arr = arr_feat[idx_pair[0] : idx_pair[1]]
            hf["data"][idx] = np.ravel(arr)


@logging_utils.timer("Plasticc Training data Processing")
def make_dataset(settings):
    """
    """

    df = pd.read_csv(os.path.join(settings.raw_dir, "training_set.csv"))
    df_meta = pd.read_csv(os.path.join(settings.raw_dir, "training_set_metadata.csv"))
    df = df.merge(df_meta, on="object_id", how="left")

    df = preprocess_plasticc(settings, df)
    df = pivot_dataframe(df)
    save_to_HDF5(settings, df)


def process_test_chunk(filename, settings):

    df = pd.read_pickle(filename)

    if len(df) == 0:
        return

    # Apply processing
    hdf5_file_name = Path(settings.processed_dir) / Path(
        os.path.basename(filename)
    ).with_suffix(".h5")
    df = preprocess_plasticc(settings, df)
    df = pivot_dataframe(df)
    save_to_HDF5(settings, df, hdf5_file_name, verbose=False)


@logging_utils.timer("Plasticc Test data Processing")
def make_test_dataset(settings):

    # Clear preproc dir
    for f in glob.glob(settings.preprocessed_dir + "/*"):
        logging_utils.print_green("Removing", f)
        os.remove(f)

    # Clear data dir of its test set constituents
    for f in glob.glob(settings.processed_dir + "/test_set_*.h5"):
        logging_utils.print_green("Removing", f)
        os.remove(f)

    num_chunks = 100

    # Load metadata
    df_meta = pd.read_csv(os.path.join(settings.raw_dir, "test_set_metadata.csv"))

    # We first cache the test dataset in multiple sub-chunks with pickle
    # This is reasonably fast

    # Count number of lines in the test set
    num_lines = int(
        subprocess.check_output(
            shlex.split(f"sed -n '$=' {os.path.join(settings.raw_dir, 'test_set.csv')}")
        )
        .decode()
        .rstrip()
    )
    chunksize = num_lines // num_chunks
    reader = pd.read_csv(
        os.path.join(settings.raw_dir, "test_set.csv"), sep=",", chunksize=chunksize
    )
    # Get the first chunk
    df_temp = next(reader)

    with tqdm(total=num_chunks, desc="Caching test data", ncols=100) as pbar:
        for chunk_idx, df_chunk in enumerate(reader):
            # Find out if df_chunk has an overlap with df_temp

            last_id = df_temp.object_id.values[-1]
            first_id = df_chunk.object_id.values[0]

            if last_id == first_id:
                df_temp = pd.concat(
                    [df_temp, df_chunk[df_chunk.object_id == last_id]]
                ).reset_index(drop=True)

            df_temp.merge(df_meta, on="object_id", how="left").to_pickle(
                os.path.join(
                    settings.preprocessed_dir, f"test_set_chunk_{chunk_idx}.pickle"
                )
            )

            df_temp = df_chunk[df_chunk.object_id != last_id].reset_index(drop=True)

            pbar.update(1)

    # Dump the last chunk
    if len(df_temp) != 0:
        df_temp.merge(df_meta, on="object_id", how="left").to_pickle(
            os.path.join(
                settings.preprocessed_dir, f"test_set_chunk_{chunk_idx + 1}.pickle"
            )
        )

    # Now we process all of the chunks in parallel
    list_pickles = glob.glob(
        os.path.join(settings.preprocessed_dir, "test_set_*.pickle")
    )

    process_fn = partial(process_test_chunk, settings=settings)

    # Split list pickles into chunks to get a progress bar with tqdm
    num_elem = len(list_pickles)
    chunk_size = min(10, num_elem)
    num_chunks = num_elem / chunk_size
    list_chunks = np.array_split(np.arange(num_elem), num_chunks)
    max_workers = multiprocessing.cpu_count() - 1
    for chunk_idxs in tqdm(list_chunks, desc="Processing test data", ncols=100):

        start_idx, end_idx = chunk_idxs[0], chunk_idxs[-1] + 1

        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            executor.map(process_fn, list_pickles[start_idx:end_idx])

    list_SNID = []
    for f in glob.glob(os.path.join(settings.processed_dir, "test_set*.h5")):
        with h5py.File(f, "r") as hf:
            list_SNID += np.ravel(hf["SNID"][:]).tolist()

    print(len(np.unique(list_SNID)))
    print(len(np.unique(df_meta.object_id.values)))



import os
import shlex
import shutil
import subprocess
from contextlib import ContextDecorator


class testmanager(ContextDecorator):

    def __enter__(self):

        self.dir_path = os.path.dirname(os.path.realpath(__file__))

        for folder in ["predictions", "preprocessed", "processed", "models"]:
            shutil.rmtree(
                os.path.join(self.dir_path, "dump", folder), ignore_errors=True
            )

        cmd = f"python run.py --data --dump_dir tests/dump"
        subprocess.check_call(shlex.split(cmd))

    def __exit__(self, type, value, traceback):

        for folder in ["predictions", "preprocessed", "processed"]:
            shutil.rmtree(
                os.path.join(self.dir_path, "dump", folder), ignore_errors=True
            )


def call_cmd(cmd):

    try:
        subprocess.check_call(shlex.split(cmd))
    except subprocess.CalledProcessError:
        assert False



import re
import os
import h5py
import torch
import numpy as np
import pandas as pd
import torch.nn as nn
from tqdm import tqdm
from pathlib import Path
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from ..utils import data_utils as du
from ..utils import logging_utils as lu
from ..utils import training_utils as tu
from ..utils.visualization_utils import FILTER_COLORS, ALL_COLORS, LINE_STYLE

plt.switch_backend("agg")


def get_predictions(settings, dict_rnn, X, target, OOD=None):

    list_data = [(X.copy(), target)]

    _, X_tensor, *_ = tu.get_data_batch(list_data, [0], settings, OOD=OOD)

    if settings.use_cuda:
        X_tensor.cuda()

    seq_len = X_tensor.shape[0]
    d_pred = {key: {"prob": []} for key in dict_rnn}

    # Loop over light curve time steps to obtain prediction for each time step
    for i in range(1, seq_len + 1):
        # Slice along the time step dimension
        X_slice = X_tensor[:i]

        # Apply rnn to obtain prediction
        for model_type, rnn in dict_rnn.items():

            n = settings.num_inference_samples if "variational" in model_type else 1
            new_size = (X_slice.size(0), n, X_slice.size(2))

            if "bayesian" in model_type:

                # Loop over num samples to obtain predictions
                list_out = [
                    rnn(X_slice.expand(new_size))
                    for i in range(settings.num_inference_samples)
                ]
                out = torch.cat(list_out, dim=0)
                # Apply softmax to obtain a proba
                pred_proba = nn.functional.softmax(out, dim=-1).data.cpu().numpy()
            else:
                out = rnn(X_slice.expand(new_size))
                # Apply softmax to obtain a proba
                pred_proba = nn.functional.softmax(out, dim=-1).data.cpu().numpy()

            # Add to buffer list
            d_pred[model_type]["prob"].append(pred_proba)

    # Stack
    for key in dict_rnn.keys():
        arr_proba = np.stack(d_pred[key]["prob"], axis=0)
        d_pred[key]["prob"] = arr_proba  # arr_prob is (T, num_samples, 2)
        d_pred[key]["median"] = np.median(arr_proba, axis=1)
        d_pred[key]["perc_16"] = np.percentile(arr_proba, 16, axis=1)
        d_pred[key]["perc_84"] = np.percentile(arr_proba, 84, axis=1)
        d_pred[key]["perc_2"] = np.percentile(arr_proba, 2, axis=1)
        d_pred[key]["perc_98"] = np.percentile(arr_proba, 98, axis=1)

    return d_pred, X_tensor.squeeze().detach().cpu().numpy()


def plot_predictions(
    settings, d_plot, SNID, redshift, peak_MJD, target, arr_time, d_pred, OOD
):

    plt.figure()
    gs = gridspec.GridSpec(2, 1)
    # Plot the lightcurve
    ax = plt.subplot(gs[0])
    for flt in d_plot.keys():
        flt_time = d_plot[flt]["MJD"]
        # Only plot a time series if it's non empty
        if len(flt_time) > 0:
            flux = d_plot[flt]["FLUXCAL"]
            fluxerr = d_plot[flt]["FLUXCALERR"]
            ax.errorbar(
                flt_time,
                flux,
                yerr=fluxerr,
                fmt="o",
                label=f"Filter {flt}",
                color=FILTER_COLORS[flt],
            )
    ax.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)
    ax.set_ylabel("FLUXCAL")
    ylim = ax.get_ylim()

    SNtype = du.sntype_decoded(target, settings)
    if settings.data_testing:
        ax.set_title(f"ID: {SNID}")
    elif OOD is not None:
        ax.set_title(f"OOD {OOD} ID: {SNID}")
    else:
        ax.set_title(SNtype + f" (ID: {SNID}, redshift: {redshift:.3g})")
        # Add PEAKMJD
        if OOD is None and not settings.data_testing and arr_time.min()<peak_MJD and peak_MJD>arr_time.max():
            ax.plot([peak_MJD, peak_MJD], ylim, "k--", label="Peak MJD")

    # Plot the classifications
    ax = plt.subplot(gs[1])
    ax.set_ylim(0, 1)

    for idx, key in enumerate(d_pred.keys()):

        for class_prob in range(settings.nb_classes):
            color = ALL_COLORS[class_prob + idx * settings.nb_classes]
            linestyle = LINE_STYLE[class_prob]
            label = du.sntype_decoded(class_prob, settings)

            if len(d_pred) > 1:
                label += f" {key}"

            ax.plot(
                arr_time,
                d_pred[key]["median"][:, class_prob],
                color=color,
                linestyle=linestyle,
                label=label,
            )
            ax.fill_between(
                arr_time,
                d_pred[key]["perc_16"][:, class_prob],
                d_pred[key]["perc_84"][:, class_prob],
                color=color,
                alpha=0.4,
            )
            ax.fill_between(
                arr_time,
                d_pred[key]["perc_2"][:, class_prob],
                d_pred[key]["perc_98"][:, class_prob],
                color=color,
                alpha=0.2,
            )

    ax.set_xlabel("Time (MJD)")
    ax.set_ylabel("classification probability")
    # Add PEAKMJD
    if OOD is None and not settings.data_testing and arr_time.min()<peak_MJD and peak_MJD>arr_time.max():
        ax.plot([peak_MJD, peak_MJD], [0, 1], "k--", label="Peak MJD")
    ax.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)

    prefix = f"OOD_{OOD}_" if OOD is not None else ""

    if len(settings.model_files) > 1:
        fig_path = f"{settings.figures_dir}/{prefix}multi_model_early_prediction"
        fig_name = f"{prefix}multi_model_{SNID}.png"
    elif len([settings.model_files]) == 1:
        parent_dir = Path(settings.model_files[0]).parent.name
        fig_path = f"{settings.lightcurves_dir}/{parent_dir}/{prefix}early_prediction"
        fig_name = f"{parent_dir}_{prefix}class_pred_with_lc_{SNID}.png"
    else:
        fig_path = (
            f"{settings.lightcurves_dir}/{settings.pytorch_model_name}/{prefix}early_prediction"
        )
        fig_name = (
            f"{settings.pytorch_model_name}_{prefix}class_pred_with_lc_{SNID}.png"
        )
    Path(fig_path).mkdir(parents=True, exist_ok=True)
    plt.tight_layout()
    plt.savefig(Path(fig_path) / fig_name)
    plt.clf()
    plt.close()


def make_early_prediction(settings, nb_lcs=1, do_gifs=False):
    """Load model corresponding to settings
    or (if specified) load a list of models.

    - Show evolution of classification for one time-step, then 2, up to all of the lightcurve
    - For Bayesian models, show uncertainty in the prediction
    - Figures are save in the figures repository

    Args:
        settings: (ExperimentSettings) custom class to hold hyperparameters
        int (nb_lcs): number of light-curves to plot, default is 1
    """
    settings.random_length = False
    settings.random_redshift = False

    # Load the test data
    list_data_test = tu.load_HDF5(settings, test=True)

    # Load features list
    file_name = f"{settings.processed_dir}/database.h5"
    with h5py.File(file_name, "r") as hf:
        features = hf["features"][settings.idx_features]

    # Load RNN model
    dict_rnn = {}
    if settings.model_files is None:
        settings.model_files = [f"{settings.rnn_dir}/{settings.pytorch_model_name}.pt"]
    else:
        # check if the model files are there
        tmp_not_found = [m for m in settings.model_files if not os.path.exists(m)]
        if len(tmp_not_found) > 0:
            print(lu.str_to_redstr(f"Files not found {tmp_not_found}"))
            tmp_model_files = [m for m in settings.model_files if os.path.exists(m)]
            settings.model_files = tmp_model_files

    for model_file in settings.model_files:
        if "variational" in model_file:
            settings.model = "variational"
        if "vanilla" in model_file:
            settings.model = "vanilla"
        if "bayesian" in model_file:
            settings.model = "bayesian"
        rnn = tu.get_model(settings, len(settings.training_features))
        rnn_state = torch.load(model_file, map_location=lambda storage, loc: storage)
        rnn.load_state_dict(rnn_state)
        rnn.to(settings.device)
        rnn.eval()
        name = (
            f"{settings.model} photometry"
            if "photometry" in model_file
            else f"{settings.model} salt"
        )
        dict_rnn[name] = rnn

    # load SN info
    SNinfo_df = du.load_HDF5_SNinfo(settings)

    # Loop over data to plot prediction
    # randomly select lcs to plot
    list_entries = np.random.randint(0,high=len(list_data_test),size=nb_lcs)
    subset_to_plot = [list_data_test[i] for i in list_entries]
    for X, target, SNID, _, X_ori in tqdm(subset_to_plot, ncols=100):
        
        try:
            redshift = SNinfo_df[SNinfo_df["SNID"] == SNID]["SIM_REDSHIFT_CMB"].values[0]
            peak_MJD = SNinfo_df[SNinfo_df["SNID"] == SNID]["PEAKMJDNORM"].values[0]
        except Exception:
            redshift = 0.0
            peak_MJD = 0.0

        # Prepare plotting data in a dict
        d_plot = {
            flt: {"FLUXCAL": [], "FLUXCALERR": [], "MJD": []}
            for flt in settings.list_filters
        }
        for OOD in [None] + du.OOD_TYPES:
            with torch.no_grad():
                d_pred, X_normed = get_predictions(
                    settings, dict_rnn, X, target, OOD=OOD
                )
            # X here has been normalized. We unnormalize X
            X_unnormed = tu.unnormalize_arr(X_normed, settings)
            # Check we do recover X_ori when OOD is None
            if OOD is None and settings.norm != 'cosmo':
                #check if normalization converges
                # using clipping in case of min<model_min
                X_clip = X_ori.copy()
                X_clip = np.clip(X_clip[:,settings.idx_features_to_normalize], settings.arr_norm[:, 0], np.inf)
                X_ori[:,settings.idx_features_to_normalize] = X_clip
                assert np.all(np.all(np.isclose(np.ravel(X_ori), np.ravel(X_unnormed), atol=1e-1)))

            # TODO: IMPROVE
            df_temp = pd.DataFrame(data=X_unnormed, columns=features)
            arr_time = np.cumsum(df_temp.delta_time.values)
            df_temp['time'] = arr_time
            for flt in settings.list_filters:
                non_zero = np.where(
                    ~np.isclose(df_temp[f"FLUXCAL_{flt}"].values, 0, atol=1E-2)
                )[0]
                d_plot[flt]["FLUXCAL"] = df_temp[f"FLUXCAL_{flt}"].values[non_zero]
                d_plot[flt]["FLUXCALERR"] = df_temp[f"FLUXCALERR_{flt}"].values[
                    non_zero
                ]
                d_plot[flt]["MJD"] = arr_time[non_zero]
            plot_predictions(
                settings,
                d_plot,
                SNID,
                redshift,
                peak_MJD,
                target,
                arr_time,
                d_pred,
                OOD,
            )

            # use to create GIFs
            if not OOD:
                if do_gifs:
                    plot_gif(settings,
                             df_temp,
                             SNID,
                             redshift,
                             peak_MJD,
                             target,
                             arr_time,
                             d_pred)
    lu.print_green("Finished plotting lightcurves and predictions ")


def plot_gif(settings, df_plot, SNID, redshift, peak_MJD, target, arr_time, d_pred
             ):
    """ Create GIFs for classification
    """
    import imageio

    def plot_image_for_gif(fig,gs, df_plot,d_pred,time,SNtype):

        # Plot the lightcurve
        ax = plt.subplot(gs[0])
        # Used to keep the limits constant
        flux_max = max(df_plot[[k for k in df_plot.keys() if 'FLUXCAL_' in k]].max())
        flux_min = min(df_plot[[k for k in df_plot.keys() if 'FLUXCAL_' in k]].min())
        ax.set_ylim(flux_min - 5, flux_max + 5)
        ax.set_xlim(-.5, max(df_plot['time']) + 2)

        # slice for gif
        df_sel = df_plot[df_plot['time'] <= time]
        for flt in settings.list_filters:
            ax.errorbar(df_sel['time'], df_sel[f"FLUXCAL_{flt}"],
                        yerr=df_sel[f"FLUXCALERR_{flt}"],
                        fmt="o",
                        label=f"Filter {flt}",
                        color=FILTER_COLORS[flt])

        ax.set(xlabel='', ylabel='flux',
               title=f"{SNtype} (ID: {SNID}, redshift: {redshift:.3g})")

        # Plot the classifications
        ax = plt.subplot(gs[1])
        ax.clear()
        ax.set_ylim(0, 1)
        ax.set_xlim(-.5, max(df_plot['time']) + 2)
        # select classification of same length
        for idx, key in enumerate(d_pred.keys()):

            for class_prob in range(settings.nb_classes):
                color = ALL_COLORS[class_prob + idx * settings.nb_classes]
                linestyle = LINE_STYLE[class_prob]
                label = du.sntype_decoded(class_prob, settings)

                if len(d_pred) > 1:
                    label += f" {key}"

                ax.plot(
                    arr_time[:len(df_sel)],
                    d_pred[key]["median"][:, class_prob][:len(df_sel)],
                    color=color,
                    linestyle=linestyle,
                    label=label,
                )
                ax.fill_between(
                    arr_time[:len(df_sel)],
                    d_pred[key]["perc_16"][:, class_prob][:len(df_sel)],
                    d_pred[key]["perc_84"][:, class_prob][:len(df_sel)],
                    color=color,
                    alpha=0.4,
                )
                ax.fill_between(
                    arr_time[:len(df_sel)],
                    d_pred[key]["perc_2"][:, class_prob][:len(df_sel)],
                    d_pred[key]["perc_98"][:, class_prob][:len(df_sel)],
                    color=color,
                    alpha=0.2,
                )
        ax.set_ylabel("classification probability")
        ax.set_xlabel("time")

        # Used to return the plot as an image rray
        fig.canvas.draw()       # draw the canvas, cache the renderer
        image = np.frombuffer(fig.canvas.tostring_rgb(), dtype='uint8')
        image = image.reshape(fig.canvas.get_width_height()[::-1] + (3,))

        from PIL import Image
        import PIL.ImageOps as pops
        im = Image.fromarray(image)
        # make background transparent
        # im = im.convert('RGB')
        # im =pops.invert(im)
        image = im

        return image

    kwargs_write = {'fps':1.0, 'quantizer':'nq'}
    fig = plt.figure()
    gs = gridspec.GridSpec(2, 1)
    SNtype = du.sntype_decoded(target, settings)

    fig_path = (
        f"{settings.lightcurves_dir}/{settings.pytorch_model_name}/gif"
    )
    fig_name = (
        f"{settings.pytorch_model_name}_class_pred_with_lc_{SNID}.gif"
    )
    Path(fig_path).mkdir(parents=True, exist_ok=True)
    arr_images = [plot_image_for_gif(fig, gs,df_plot,d_pred,time,SNtype) for time in arr_time]
    arr_images[0].save(str(Path(fig_path) / fig_name), save_all=True, append_images=arr_images, loop=5, duration=200)



import numpy as np
from sklearn.ensemble import RandomForestClassifier

from ..utils import data_utils as du
from ..utils import training_utils as tu
from ..utils import logging_utils


def train(settings):
    """Train random forest models

    - Features are stored in a .FITRES file found in settings.processed_dir.
    - Carry out a train/val split based on predefined splits
    - Compute evaluation metrics on train and validation set
    - Save trained models to settings.models_dir

    Args:
        settings: (ExperimentSettings) custom class to hold hyperparameters
    """

    assert settings.source_data == "saltfit", logging_utils.str_to_redstr(
        "Only salfit is a valid data source for random forest"
    )
    assert settings.nb_classes == 2, logging_utils.str_to_redstr(
        "Binary classification is the only task allowed for RandomForest"
    )

    clf = RandomForestClassifier(
        bootstrap=settings.bootstrap,
        min_samples_leaf=settings.min_samples_leaf,
        n_estimators=settings.n_estimators,
        min_samples_split=settings.min_samples_split,
        criterion=settings.criterion,
        max_features=settings.max_features,
        max_depth=settings.max_depth,
        random_state=settings.seed,
        n_jobs=-1,
    )

    ###################
    # Data
    ###################

    # load data
    df_data = du.load_fitfile(settings)
    SNID = df_data["SNID"].values

    # Load training and validation SNID to make our train/val split
    sn_df = du.load_HDF5_SNinfo(settings)
    df_train = sn_df[sn_df["dataset_photometry_2classes"] == 0]
    n_train = len(df_train)
    df_train = df_train[: int(settings.data_fraction * n_train)]

    df_val = sn_df[sn_df["dataset_photometry_2classes"] == 1]
    n_val = len(df_val)
    df_val = df_val[: int(settings.data_fraction * n_val)]

    # Compute train/val split index
    idx_train = np.where(np.in1d(SNID, df_train.SNID.values))[0]
    idx_val = np.where(np.in1d(SNID, df_val.SNID.values))[0]

    # Add redshift features if required by settings
    df_data = du.add_redshift_features(settings, df_data)

    ###################
    # Training
    ###################

    # Prepare features and target
    X = df_data[settings.randomforest_features].values
    y = df_data["target_2classes"].values

    logging_utils.print_green("Features", ",".join(settings.randomforest_features))

    # Apply train/val split
    X_train, X_val = X[idx_train], X[idx_val]
    y_train, y_val = y[idx_train], y[idx_val]

    # Fit  and evaluate model
    clf = tu.train_and_evaluate_randomforest_model(clf, X_train, y_train, X_val, y_val)
    # save the model to disk
    tu.save_randomforest_model(settings, clf)



import os
import json
import h5py
import itertools
import numpy as np
from pathlib import Path
from collections import OrderedDict
from .data_utils import PLASTICC_FILTERS


class ExperimentSettings:
    """Mother class to control experiment parameters

    This class is responsible for the following

    - Defining paths and model names
    - Choosing the device on which to run computations
    - Specifying all hyperparameters such as model configuration, datasets, features etc

    Args:
        cli_args (argparse.Namespace) command line arguments
    """

    def __init__(self, cli_args):

        # Transfer attributes
        if isinstance(cli_args, dict):
            self.__dict__.update(cli_args)
            self.cli_args = cli_args
        else:
            self.__dict__.update(cli_args.__dict__)
            self.cli_args = cli_args.__dict__

        self.device = "cpu"
        if self.use_cuda:
            self.device = "cuda"

        if self.model == "variational":
            self.weight_decay = self.weight_decay
        else:
            self.weight_decay = 0.0

        # Load simulation and training settings and prepare directories
        self.setup_dir()

        self.set_pytorch_model_name()
        self.set_randomforest_model_name()

        self.randomforest_features = self.get_randomforest_features()

        # Set the database file names
        self.set_database_file_names()

        # Set the feature lists
        self.set_feature_lists()

        # Get the feature normalization dict
        self.load_normalization()

        self.overwrite = not self.no_overwrite

        # filter combination
        list_filters_combination = []
        for i in range(1, len(self.list_filters)+1):
            tmp = [''.join(t) for t in list(
                itertools.combinations(self.list_filters, i))]
            list_filters_combination = list_filters_combination + tmp
        self.list_filters_combination = list_filters_combination

    def get_randomforest_features(self):
        """Specify list of features to be used for RandomForest training
        """

        features = [
            "x1",
            "x1ERR",
            "c",
            "cERR",
            "mB",
            "mBERR",
            "x0",
            "x0ERR",  # 'COV_x1_c', 'COV_x1_x0','COV_c_x0', 'NDOF',
            "FITCHI2",
            "m0obs_r",
            "m0obs_i",
            "m0obs_g",
            "m0obs_z",
            "em0obs_i",
            "em0obs_r",
            "em0obs_g",
            "em0obs_z",
        ]

        if self.redshift == "zpho":
            features += ["HOSTGAL_PHOTOZ", "HOSTGAL_PHOTOZ_ERR"]
        elif self.redshift == "zspe":
            features += ["HOSTGAL_SPECZ", "HOSTGAL_SPECZ_ERR"]

        return features

    def setup_dir(self):
        """Configure directories where data is read from or dumped to
        during the course of an experiment
        """

        for path in [
            f"{self.raw_dir}",
            f"{self.fits_dir}",
            f"{self.dump_dir}/explore",
            f"{self.dump_dir}/stats",
            f"{self.dump_dir}/figures",
            f"{self.dump_dir}/lightcurves",
            f"{self.dump_dir}/latex",
            f"{self.dump_dir}/processed",
            f"{self.dump_dir}/preprocessed",
            f"{self.dump_dir}/models",
        ]:

            setattr(self, Path(path).name + "_dir", path)

            Path(path).mkdir(exist_ok=True, parents=True)

    def set_pytorch_model_name(self):
        """Define the model name for all NN based classifiers
        """
        name = f"{self.model}_S_{self.seed}_CLF_{self.nb_classes}"
        name += f"_R_{self.redshift}"
        name += f"_{self.source_data}_DF_{self.data_fraction}_N_{self.norm}"
        name += f"_{self.layer_type}_{self.hidden_dim}x{self.num_layers}"
        name += f"_{self.dropout}"
        name += f"_{self.batch_size}"
        name += f"_{self.bidirectional}"
        name += f"_{self.rnn_output_option}"
        if "bayesian" in self.model:
            name += (
                f"_Bayes_{self.pi}_{self.log_sigma1}_{self.log_sigma2}"
                f"_{self.rho_scale_lower}_{self.rho_scale_upper}"
                f"_{self.log_sigma1_output}_{self.log_sigma2_output}"
                f"_{self.rho_scale_lower_output}_{self.rho_scale_upper_output}"
            )
        if self.cyclic:
            name += "_C"
        if self.weight_decay > 0:
            name += f"_WD_{self.weight_decay}"

        self.pytorch_model_name = name
        self.rnn_dir = f"{self.models_dir}/{self.pytorch_model_name}"

        if self.train_rnn:
            os.makedirs(self.rnn_dir, exist_ok=True)
            # Dump the command line arguments (for model restoration)
            with open(Path(self.rnn_dir) / "cli_args.json", "w") as f:
                json.dump(self.cli_args, f, indent=4, sort_keys=True)

    def set_randomforest_model_name(self):
        """Define the model name for all RandomForest based classifiers
        """

        name = f"randomforest_S_{self.seed}_CLF_{self.nb_classes}"
        name += f"_R_{self.redshift}"
        name += f"_{self.source_data}_DF_{self.data_fraction}_N_{self.norm}"

        self.randomforest_model_name = name
        self.rf_dir = f"{self.models_dir}/{self.randomforest_model_name}"

        if self.train_rf:
            os.makedirs(self.rf_dir, exist_ok=True)
            # Dump the command line arguments (for model restoration)
            with open(Path(self.rf_dir) / "cli_args.json", "w") as f:
                json.dump(self.cli_args, f, indent=4, sort_keys=True)

        return name

    def check_data_exists(self):
        """Utility to check the database has been built
        """

        database_file = f"{self.processed_dir}/database.h5"
        assert os.path.isfile(database_file)

    def set_feature_lists(self):
        """Utility to define the features used to train NN=based models
        """

        self.training_features_to_normalize = [
            f"FLUXCAL_{f}" for f in self.list_filters
        ]
        self.training_features_to_normalize += [
            f"FLUXCALERR_{f}" for f in self.list_filters
        ]
        self.training_features_to_normalize += ["delta_time"]

        if not self.data:
            # If the database has been created, add the list of all features
            with h5py.File(self.hdf5_file_name, "r") as hf:
                self.all_features = hf["features"][:].astype(str)

                self.non_redshift_features = [
                    f for f in self.all_features if "HOSTGAL" not in f
                ]

                # Optionally add redshift
                self.redshift_features = []
                if self.redshift == "zpho":
                    self.redshift_features = [
                        f for f in self.all_features if "HOSTGAL_PHOTOZ" in f
                    ]
                elif self.redshift == "zspe":
                    self.redshift_features = [
                        f for f in self.all_features if "HOSTGAL_SPECZ" in f
                    ]

                self.training_features = (
                    self.non_redshift_features + self.redshift_features
                )

    def set_database_file_names(self):
        """Create a unique database name based on the dataset required
        by the settings
        """

        out_file = f"{self.processed_dir}/database"
        self.pickle_file_name = out_file + ".pickle"
        self.hdf5_file_name = out_file + ".h5"

    def load_normalization(self):
        """Create an array holding the data-normalization parameters
        used to normalize certain features in the NN-based classification
        pipeline
        """

        if not self.data:

            self.idx_features = [
                i
                for (i, f) in enumerate(self.all_features)
                if f in self.training_features
            ]

            self.idx_specz = [
                i
                for (i, f) in enumerate(self.training_features)
                if "HOSTGAL_SPECZ" in f
            ]

            self.idx_flux = [
                i for (i, f) in enumerate(self.training_features) if "FLUXCAL_" in f
            ]

            self.idx_fluxerr = [
                i for (i, f) in enumerate(self.training_features) if "FLUXCALERR_" in f
            ]

            self.idx_delta_time = [
                i for (i, f) in enumerate(self.training_features) if "delta_time" in f
            ]

            self.idx_features_to_normalize = [
                i
                for (i, f) in enumerate(self.all_features)
                if f in self.training_features_to_normalize
            ]

            self.d_feat_to_idx = {f: i for i,
                                  f in enumerate(self.all_features)}

            list_norm = []

            with h5py.File(self.hdf5_file_name, "r") as hf:

                for f in self.training_features_to_normalize:

                    if self.norm == "perfilter":

                        minv = np.array(hf[f"normalizations/{f}/min"])
                        meanv = np.array(hf[f"normalizations/{f}/mean"])
                        stdv = np.array(hf[f"normalizations/{f}/std"])

                        list_norm.append([minv, meanv, stdv])

                    else:

                        if "FLUX" in f:
                            prefix = f.split("_")[0]
                            minv = np.array(hf[f"normalizations_global/{prefix}/min"])
                            meanv = np.array(hf[f"normalizations_global/{prefix}/mean"])
                            stdv = np.array(hf[f"normalizations_global/{prefix}/std"])
                        else:
                            minv = np.array(hf[f"normalizations/{f}/min"])
                            meanv = np.array(hf[f"normalizations/{f}/mean"])
                            stdv = np.array(hf[f"normalizations/{f}/std"])

                        list_norm.append([minv, meanv, stdv])

            self.arr_norm = np.array(list_norm)


class PlasticcSettings(object):

    def __init__(self, cli_args):

        # Transfer attributes
        if isinstance(cli_args, dict):
            self.__dict__.update(cli_args)
        else:
            self.__dict__.update(cli_args.__dict__)

        self.dump_dir = os.path.join(self.dump_dir, "plasticc")
        self.nb_classes = 14

        self.device = "cpu"
        if self.use_cuda:
            self.device = "cuda"

        if self.model == "variational":
            self.weight_decay = self.weight_decay
        else:
            self.weight_decay = 0.0

        # Load simulation and training settings and prepare directories
        self.setup_dir()

        self.pytorch_model_name = self.get_pytorch_model_name()

        # Set the filters used in the study
        self.list_filters = PLASTICC_FILTERS

        # Set the database file names
        self.set_database_file_names()

        # Set the feature lists
        self.set_feature_lists()

        # Get the feature normalization dict
        self.load_normalization()

        # TODO
        # Set SN types for multiclass classification

        self.overwrite = not self.no_overwrite

    def check_data_exists(self):

        database_file = f"{self.processed_dir}/database.h5"
        assert os.path.isfile(database_file)

    def setup_dir(self):

        # Check raw data exists in the dump_dir
        assert os.path.exists(self.raw_dir)

        for path in [
            f"{self.dump_dir}/explore",
            f"{self.dump_dir}/stats",
            f"{self.dump_dir}/figures",
            f"{self.dump_dir}/latex",
            f"{self.dump_dir}/processed",
            f"{self.dump_dir}/preprocessed",
            f"{self.dump_dir}/models",
        ]:

            setattr(self, Path(path).name, path)

            Path(path).mkdir(exist_ok=True, parents=True)

    def get_pytorch_model_name(self):

        name = f"{self.model}_S_{self.seed}_CLF_{self.nb_classes}"
        name += f"_R_{self.redshift}"
        name += f"_{self.source_data}_DF_{self.data_fraction}_N_{self.norm}"
        name += f"_{self.layer_type}_{self.hidden_dim}x{self.num_layers}"
        name += f"_{self.dropout}"
        name += f"_{self.batch_size}"
        name += f"_{self.bidirectional}"
        name += f"_{self.rnn_output_option}"
        if "bayesian" in self.model:
            name += (
                f"_Bayes_{self.pi}_{self.log_sigma1}_{self.log_sigma2}_{self.rho_scale_lower}_{self.rho_scale_upper}"
                f"_{self.log_sigma1_output}_{self.log_sigma2_output}_{self.rho_scale_lower_output}_{self.rho_scale_upper_output}"
            )
        if self.cyclic:
            name += "_C"
        if self.weight_decay > 0:
            name += f"_WD_{self.weight_decay}"

        if self.train_plasticc:
            os.makedirs(os.path.join(self.models_dir, name), exist_ok=True)

        return name

    def set_feature_lists(self):

        self.training_features_to_normalize = [
            f"FLUXCAL_{f}" for f in self.list_filters
        ]
        self.training_features_to_normalize += [
            f"FLUXCALERR_{f}" for f in self.list_filters
        ]
        self.training_features_to_normalize += ["delta_time"]

        if self.train_plasticc or self.predict_plasticc:
            # If the database has been created, add the list of all features
            with h5py.File(self.hdf5_file_name, "r") as hf:
                self.all_features = hf["features"][:].astype(str)

                # Optionally add redshift
                self.redshift_features = []
                if self.redshift == "zpho":
                    # use only photometric redshift features
                    self.redshift_features = [
                        f for f in self.all_features if "HOSTGAL_PHOTOZ" in f
                    ]
                elif self.redshift == "zspe":
                    # Use all redshift features
                    self.redshift_features = [
                        f for f in self.all_features if "HOST" in f
                    ]

                self.non_redshift_features = [
                    f for f in self.all_features if "HOST" not in f
                ]

                self.training_features = (
                    self.non_redshift_features + self.redshift_features
                )
                self.training_features = [
                    f for f in self.all_features if f in self.training_features
                ]

    def set_database_file_names(self):
        out_file = f"{self.processed_dir}/database"
        self.pickle_file_name = out_file + ".pickle"
        self.hdf5_file_name = out_file + ".h5"

    def load_normalization(self):

        if self.train_plasticc or self.predict_plasticc:

            self.idx_features = [
                i
                for (i, f) in enumerate(self.all_features)
                if f in self.training_features
            ]

            self.idx_specz = [
                i
                for (i, f) in enumerate(self.training_features)
                if "HOSTGAL_SPECZ" in f
            ]

            self.idx_features_to_normalize = [
                i
                for (i, f) in enumerate(self.all_features)
                if f in self.training_features_to_normalize
            ]

            self.d_feat_to_idx = {f: i for i,
                                  f in enumerate(self.all_features)}

            list_norm = []

            with h5py.File(self.hdf5_file_name, "r") as hf:

                for f in self.training_features_to_normalize:

                    if self.norm == "perfilter":

                        minv = np.array(hf[f"normalizations/{f}/min"])
                        meanv = np.array(hf[f"normalizations/{f}/mean"])
                        stdv = np.array(hf[f"normalizations/{f}/std"])

                        list_norm.append([minv, meanv, stdv])

                    else:

                        if "FLUX" in f:
                            prefix = f.split("_")[0]
                            minv = np.array(hf[f"normalizations_global/{prefix}/min"])
                            meanv = np.array(hf[f"normalizations_global/{prefix}/mean"])
                            stdv = np.array(hf[f"normalizations_global/{prefix}/std"])
                        else:
                            minv = np.array(hf[f"normalizations/{f}/min"])
                            meanv = np.array(hf[f"normalizations/{f}/mean"])
                            stdv = np.array(hf[f"normalizations/{f}/std"])

                        list_norm.append([minv, meanv, stdv])

            self.arr_norm = np.array(list_norm)



import os
import glob
import h5py
import json
import numpy as np
import pandas as pd
from tqdm import tqdm
from pathlib import Path
from natsort import natsorted
from astropy.table import Table
from collections import namedtuple, OrderedDict

from . import logging_utils

OFFSETS = [-2, -1, 0, 1, 2]
OOD_TYPES = ["random", "reverse", "shuffle", "sin"]
OFFSETS_STR = ["-2", "-1", "", "+1", "+2"]

PLASTICC_FILTERS = natsorted(["u", "g", "r", "i", "z", "y"])
DICT_PLASTICC_FILTERS = {0: "u", 1: "g", 2: "r", 3: "i", 4: "z", 5: "y"}
DICT_PLASTICC_CLASS = OrderedDict(
    {
        6: 0,
        15: 1,
        16: 2,
        42: 3,
        52: 4,
        53: 5,
        62: 6,
        64: 7,
        65: 8,
        67: 9,
        88: 10,
        90: 11,
        92: 12,
        95: 13,
        99: 14,
    }
)
LogStandardized = namedtuple(
    "LogStandardized", ["arr_min", "arr_mean", "arr_std"])


def load_pandas_from_fit(fit_file_path):
    """Load a FIT file and cast it to a PANDAS dataframe

    Args:
        fit_file_path (str): path to FIT file

    Returns:
        (pandas.DataFrame) load dataframe from FIT file
    """
    dat = Table.read(fit_file_path, format="fits")
    df = dat.to_pandas()

    return df


def sntype_decoded(target, settings):
    """Match the target class (integer in {0, ..., 6} to the name
    of the class, i.e. something like "SN Ia" or "SN CC"

    Args:
        target (int): specifies the classification target
        settings (ExperimentSettings): custom class to hold hyperparameters

    Returns:
        (str) the name of the class

    """
    if settings.nb_classes > 3:
        SNtype = list(settings.sntypes.values())[target]
    elif settings.nb_classes == 3:
        if target == 0:
            SNtype = f"SN {list(settings.sntypes.values())[0]}"
        elif target == 1:
            SNtype = f"SN CC Ix"
        else:
            SNtype = "SN CC IIx"
    else:
        list_types = list(set([x for x in settings.sntypes.values()]))
        if target == 0:
            if "Ia" in list_types:
                SNtype = "SN Ia"
            else:
                SNtype = f"SN {list(settings.sntypes.values())[0]}"
        else:
            if "Ia" in list_types:
                SNtype = f"SN {'|'.join(set([k for k in settings.sntypes.values() if 'Ia' not in k]))}"
            else:
                SNtype = f"SN {'|'.join(list(settings.sntypes.values())[1:])}"
    return SNtype


def tag_type(df, settings, type_column="TYPE"):
    """Create classes based on a type columns

    Depending on the number of classes (2 or all), we create distinct
    target columns

    Args:
        df (pandas.DataFrame): the input dataframe
        settings (ExperimentSettings): controls experiment hyperparameters
        type_column (str): the type column in df

    Returns:
        (pandas.DataFrame) the dataframe, with new target columns
    """

    # 2 classes
    # taking the first type vs. others
    list_types = list(set([x for x in settings.sntypes.values()]))
    if "Ia" in list_types:
        df[type_column] = df[type_column].astype(str)
        # get keys of Ias, the rest tag them as CC
        keys_ia = [key for (key, value)
                   in settings.sntypes.items() if value == "Ia"]
        df["target_2classes"] = df[type_column].apply(
            lambda x: 0 if x in keys_ia else 1
        )
    else:
        arr_temp = df[type_column].values.copy()
        df["target_2classes"] = (
            arr_temp != int(list(settings.sntypes.keys())[0])
        ).astype(np.uint8)

    # All classes
    # check if all types are given in input dictionary
    tmp = df[~df[type_column].isin(settings.sntypes.keys())][type_column]
    if len(tmp) > 0:
        logging_utils.print_red("Not all sntypes are given in input")
        logging_utils.print_red(f"missing: {tmp.unique()}")
        logging_utils.print_red(f"tagging them as class {len(settings.sntypes)}")

    classes_to_use = {}
    for i, k in enumerate(settings.sntypes.keys()):
        classes_to_use[k] = i
    for kk in tmp.unique():
        classes_to_use[kk] = len(settings.sntypes)

    df[f"target_{len(settings.sntypes)}classes"] = df[type_column].apply(
        lambda x: classes_to_use[x]
    )

    return df


def load_fitfile(settings, verbose=True):
    """Load the FITOPT file as a pandas dataframe

    Pickle it for future use (it is faster to load as a pickled dataframe)

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        verbose (bool): whether to display logging message. Default: ``True``

    Returns:
        (pandas.DataFrame) dataframe with FITOPT data
    """
    if verbose:
        logging_utils.print_green("Loading FITRES file")

    if Path(f"{settings.preprocessed_dir}/FITOPT000.FITRES.pickle").exists():
        df = pd.read_pickle(f"{settings.preprocessed_dir}/FITOPT000.FITRES.pickle")
        if verbose:
            print(f"Loaded {settings.preprocessed_dir}/FITOPT000.FITRES.pickle")

    elif (
        Path(f"{settings.fits_dir}/FITOPT000.FITRES").exists()
        or Path(f"{settings.fits_dir}/FITOPT000.FITRES.gz").exists()
    ):
        fit_name = (
            f"{settings.fits_dir}/FITOPT000.FITRES"
            if Path(f"{settings.fits_dir}/FITOPT000.FITRES").exists()
            else f"{settings.fits_dir}/FITOPT000.FITRES.gz"
        )
        df = pd.read_csv(
            fit_name, index_col=False, comment="#", delimiter=" ", skipinitialspace=True
        )
        df = tag_type(df, settings)

        # Rename CID to SNID
        # SNID is CID in FITOPT000.FITRES
        df = df.rename(columns={"CID": "SNID"})

        # Save to pickle for later use and fast reload
        df.to_pickle(f"{settings.preprocessed_dir}/FITOPT000.FITRES.pickle")
        if verbose:
            print(f"Loaded {fit_name}")
    else:
        # returning empty df
        df = pd.DataFrame()
        if verbose:
            logging_utils.print_yellow("Warning: No FITRES file to load")

    return df


def process_header_FITS(file_path, settings, columns=None):
    """Read the HEAD FIT file, add target columns and return
    in pandas DataFrame format

    Args:
        file_path (str): the  path to the header FIT file
        settings (ExperimentSettings): controls experiment hyperparameters
        columns (lsit): list of columns to keep. Default: ``None``

    Returns:
        (pandas.DataFrame) the dataframe, with new target columns
    """

    # Data
    df = load_pandas_from_fit(file_path)

    try:
        df['SNID'] = df['SNID'].str.decode('utf-8')
    except Exception:
        df['SNID'] = df['SNID'].astype(str)

    df = tag_type(df, settings, type_column="SNTYPE")

    if columns is not None:
        df = df[columns]

    return df


def process_header_csv(file_path, settings, columns=None):
    """Read the HEAD csv file, add target columns and return
    in pandas DataFrame format

    Args:
        file_path (str): the  path to the header FIT file
        settings (ExperimentSettings): controls experiment hyperparameters
        columns (lsit): list of columns to keep. Default: ``None``

    Returns:
        (pandas.DataFrame) the dataframe, with new target columns
    """

    # Data
    df = pd.read_csv(file_path)
    df = tag_type(df, settings, type_column="SNTYPE")

    if columns is not None:
        df = df[columns]
    return df


def add_redshift_features(settings, df):
    """Add redshift features to pandas dataframe.

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        df (str): pandas DataFrame with FIT data

    Returns:
        (pandas.DataFrame) the dataframe, possibly with added redshift features
    """

    # check if we use host redshift as feature
    host_features = [f for f in settings.randomforest_features if "HOST" in f]
    use_redshift = len(host_features) > 0

    if use_redshift > 0:
        logging_utils.print_green("Adding redshift features...")

        columns_to_read = ["SNID"] + host_features
        # reading from batch pickles
        list_files = natsorted(glob.glob(f"{settings.preprocessed_dir}/*_PHOT.pickle"))
        # Check file with redshift features exist
        error_msg = "Preprocessed_file not found. Call python run.py --data"
        assert os.path.isfile(list_files[0]), error_msg

        extra_info_df = pd.concat(
            [pd.read_pickle(f)[columns_to_read] for f in list_files]
        )

        # In extra_info_df, there are many SNID duplicates as each row corresponds to a time step in a given curve
        # We use groupby + first to only select the first row of each lightcurve
        # Then we can merge knowing there won't be SNID duplicates in extra_info_df
        extra_info_df = (
            extra_info_df.groupby("SNID")[host_features].first().reset_index()
        )

        # Add redshift info to df
        df = df.merge(extra_info_df, how="left", on="SNID")

    return df


def compute_delta_time(df):
    """Compute the delta time between two consecutive observations

    Args:
        df (pandas.DataFrame): dataframe holding lightcurve data

    Returns:
        (pandas.DataFrame) dataframe holding lightcurve data with delta_time features
    """

    df["delta_time"] = df["MJD"].diff()
    # Fill the first row with 0 to replace NaN
    df.delta_time = df.delta_time.fillna(0)
    try:
        IDs = df.SNID.values
    # Deal with the case where lightcrv_ID is the index
    except AttributeError:
        assert df.index.name == "SNID"
        IDs = df.index.values
    # Find idxs of rows where a new light curve start then zero delta_time
    idxs = np.where(IDs[:-1] != IDs[1:])[0] + 1
    arr_delta_time = df.delta_time.values
    arr_delta_time[idxs] = 0
    df["delta_time"] = arr_delta_time

    return df


def remove_data_post_large_delta_time(df):
    """
    Remove rows in the same light curve after a gap > 150 days
    Reason: If no signal has been saved in a time frame of 150 days,
    it is unlikely there is much left afterwards

    Args:
        df (pandas.DataFrame): dataframe holding lightcurve data

    Returns:
        (pandas.DataFrame) dataframe where large delta time rows have been removed
    """

    # Identify indices where delta time is large
    list_to_remove = []
    idx_high = np.where(df.delta_time.values > 150)[0]
    # Identify the lightcurve ID where this happens
    IDs = df.SNID.values
    # Loop over indices and remove row if they belong to the same light curve
    for idx in idx_high:
        ID = IDs[idx]
        same_lc = True
        while same_lc:
            list_to_remove.append(idx)
            idx += 1
            new_ID = IDs[idx]
            same_lc = new_ID == ID
    df = df.drop(list_to_remove)
    # Reset index to account for dropped rows
    df.reset_index(inplace=True, drop=True)

    return df


def load_HDF5_SNinfo(settings):
    """Load physical information related to the created database of lightcurves

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters

    Returns:
        (pandas.DataFrame) dataframe holding physics information about the dataset
    """

    file_name = f"{settings.processed_dir}/database.h5"

    dict_SNinfo = {}
    with h5py.File(file_name, "r") as hf:

        columns_to_keep = ["SNID", "SNTYPE", "mB", "c", "x1"]

        columns_to_keep += [c for c in hf.keys() if "SIM_" in c]
        columns_to_keep += [c for c in hf.keys() if "dataset_" in c]
        columns_to_keep += [c for c in hf.keys() if "PEAK" in c]

        for key in columns_to_keep:
            dict_SNinfo[key] = hf[key][:]
    df_SNinfo = pd.DataFrame(dict_SNinfo)

    return df_SNinfo


def log_standardization(arr):
    """Normalization strategy for the fluxes and fluxes error

    - Log transform the data
    - Mean and std dev normalization

    Args:
        arr (np.array): data to normalize

    Returns:
        (LogStandardized) namedtuple holding normalization data
    """

    arr_min = np.min(arr)
    arr_log = np.log(-arr_min + arr + 1e-5)
    arr_mean = arr_log.mean()
    arr_std = arr_log.std()

    return LogStandardized(arr_min=arr_min, arr_mean=arr_mean, arr_std=arr_std)


def save_to_HDF5(settings, df):
    """Saved processed dataframe to HDF5

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        df (pandas.DataFrame): dataframe holding processed data

    """
    # One hot encode filter information and Normalize features
    list_training_features = [f"FLUXCAL_{f}" for f in settings.list_filters]
    list_training_features += [f"FLUXCALERR_{f}" for f in settings.list_filters]
    list_training_features += [
        "delta_time",
        "HOSTGAL_PHOTOZ",
        "HOSTGAL_PHOTOZ_ERR",
        "HOSTGAL_SPECZ",
        "HOSTGAL_SPECZ_ERR",
    ]

    list_misc_features = [
        "PEAKMJD",
        "SNTYPE",
        "mB",
        "c",
        "x1",
        "SIM_REDSHIFT_CMB",
        "SIM_PEAKMAG_z",
        "SIM_PEAKMAG_g",
        "SIM_PEAKMAG_r",
        "SIM_PEAKMAG_i",
    ]
    list_misc_features = [k for k in list_misc_features if k in df.keys()]

    assert df.index.name == "SNID", "Must set SNID as index"

    # Get the list of lightcurve IDs
    ID = df.index.values
    # Find out when ID changes => find start and end idx of each lightcurve
    idx_change = np.where(ID[1:] != ID[:-1])[0] + 1
    idx_change = np.hstack(([0], idx_change, [len(df)]))
    list_start_end = [(s, e) for s, e in zip(idx_change[:-1], idx_change[1:])]
    # N.B. We could use df.loc[SNID], more elegant but much slower

    # Filter list start end so we get only light curves with at least 3 points
    # except when creating testing data (we want to classify all lcs even w. 1-2 epochs)
    if not settings.data_testing:
        list_start_end = list(
            filter(lambda x: x[1] - x[0] >= 3, list_start_end))

    # Shuffle
    np.random.shuffle(list_start_end)

    # Save hdf5 file
    with h5py.File(settings.hdf5_file_name, "w") as hf:

        n_samples = len(list_start_end)
        list_classes = list(set([2, len(settings.sntypes.keys())]))
        list_names = ["target", "dataset_photometry", "dataset_saltfit"]

        # These arrays can be filled in one shot
        start_idxs = [i[0] for i in list_start_end]
        shuffled_ID = ID[start_idxs]
        hf.create_dataset(
            "SNID", data=shuffled_ID, dtype=h5py.special_dtype(vlen=str)
        )
        df_SNID = pd.DataFrame(shuffled_ID, columns=["SNID"])
        logging_utils.print_green("Saving misc features")
        for feat in list_misc_features:
            if feat == "SNTYPE":
                dtype = np.dtype("int32")
            else:
                dtype = np.dtype("float32")
            hf.create_dataset(
                feat, data=df[feat].values[start_idxs], dtype=dtype)
            df.drop(columns=feat, inplace=True)

        logging_utils.print_green("Saving class")
        for c_ in list_classes:
            for name in list_names:
                field_name = f"{name}_{c_}classes"
                hf.create_dataset(
                    field_name,
                    data=df[field_name].values[start_idxs],
                    dtype=np.dtype("int8"),
                )
                df.drop(columns=field_name, inplace=True)

        df["time"] = df[["delta_time"]].groupby(df.index).cumsum()
        df = df.reset_index()

        logging_utils.print_green("Saving unique nights")
        # Compute how many unique nights of data taking existed around PEAKMJD
        for offset, suffix in zip(OFFSETS, OFFSETS_STR):
            new_column = f"PEAKMJD{suffix}_unique_nights"
            df_nights = (
                df[df["time"] < df["PEAKMJDNORM"] +
                    offset][["PEAKMJDNORM", "SNID"]]
                .groupby("SNID")
                .count()
                .astype(np.uint8)
                .rename(columns={f"PEAKMJDNORM": new_column})
                .reset_index()
            )

            hf.create_dataset(
                new_column,
                data=df_SNID.merge(df_nights, on="SNID", how="left")[
                    new_column].values,
                dtype=np.dtype("uint8"),
            )

        logging_utils.print_green("Saving filter occurences")
        # Compute how many occurences of a specific filter around PEAKMJD
        for flt in settings.list_filters:
            # Check presence / absence of the filter at all time steps
            df[f"has_{flt}"] = df.FLT.str.contains(flt).astype(np.uint8)
            for offset, suffix in zip(OFFSETS, OFFSETS_STR):
                new_column = f"PEAKMJD{suffix}_num_{flt}"
                df_flt = (
                    df[df["time"] < df["PEAKMJDNORM"] + offset][[f"has_{flt}", "SNID"]]
                    .groupby("SNID")
                    .sum()
                    .astype(np.uint8)
                    .rename(columns={f"has_{flt}": new_column})
                    .reset_index()
                )
                hf.create_dataset(
                    new_column,
                    data=df_SNID.merge(df_flt, on="SNID", how="left")[
                        new_column
                    ].values,
                    dtype=np.dtype("uint8"),
                )

            df.drop(columns=f"has_{flt}", inplace=True)

        # FInally save PEAKMJDNORM
        hf.create_dataset(
            "PEAKMJDNORM",
            data=df["PEAKMJDNORM"].values[start_idxs],
            dtype=np.dtype("float32"),
        )

        df.drop(columns=["time", "SNID", "PEAKMJDNORM"], inplace=True)

        ########################
        # Normalize per feature
        ########################
        logging_utils.print_green("Compute normalizations")
        gnorm = hf.create_group("normalizations")

        # using normalization per feature
        for feat in settings.training_features_to_normalize:
            # Log transform plus mean subtraction and standard dev subtraction
            log_standardized = log_standardization(df[feat].values)
            # Store normalization parameters
            gnorm.create_dataset(f"{feat}/min", data=log_standardized.arr_min)
            gnorm.create_dataset(f"{feat}/mean", data=log_standardized.arr_mean)
            gnorm.create_dataset(f"{feat}/std", data=log_standardized.arr_std)

        #####################################
        # Normalize flux and fluxerr globally
        #####################################
        logging_utils.print_green("Compute global normalizations")
        gnorm = hf.create_group("normalizations_global")

        ################
        # FLUX features
        #################
        flux_features = [f"FLUXCAL_{f}" for f in settings.list_filters]
        flux_log_standardized = log_standardization(df[flux_features].values)
        # Store normalization parameters
        gnorm.create_dataset(f"FLUXCAL/min", data=flux_log_standardized.arr_min)
        gnorm.create_dataset(f"FLUXCAL/mean", data=flux_log_standardized.arr_mean)
        gnorm.create_dataset(f"FLUXCAL/std", data=flux_log_standardized.arr_std)

        ###################
        # FLUXERR features
        ###################
        fluxerr_features = [f"FLUXCALERR_{f}" for f in settings.list_filters]
        fluxerr_log_standardized = log_standardization(
            df[fluxerr_features].values)
        # Store normalization parameters
        gnorm.create_dataset(f"FLUXCALERR/min", data=fluxerr_log_standardized.arr_min)
        gnorm.create_dataset(f"FLUXCALERR/mean", data=fluxerr_log_standardized.arr_mean)
        gnorm.create_dataset(f"FLUXCALERR/std", data=fluxerr_log_standardized.arr_std)

        ####################################
        # Save the rest of the data to hdf5
        ####################################

        logging_utils.print_green("Save non-data features to HDF5")

        # This type allows one to store flat arrays of variable
        # length inside an HDF5 group
        data_type = h5py.special_dtype(vlen=np.dtype("float32"))

        hf.create_dataset("data", (n_samples,), dtype=data_type)

        # If header does not have HOST info fill with empty arrays
        list_to_fill = [
            k for k in list_training_features if k not in df.columns.values.tolist()]
        if len([k for k in list_to_fill if 'HOST' not in k]) > 0:
            logging_utils.print_red("missing information in input")
            raise AttributeError
        for key in list_to_fill:
            df[key] = np.zeros(len(df))

        logging_utils.print_green("Fit onehot on FLT")
        assert sorted(df.columns.values.tolist()) == sorted(
            list_training_features + ["FLT"]
        )

        # Fit a one hot encoder for FLT
        # to have the same onehot for all datasets
        tmp = pd.Series(settings.list_filters_combination).append(df["FLT"])
        tmp_onehot = pd.get_dummies(tmp)
        # this is ok since it goes by length not by index (which I never reset)
        FLT_onehot = tmp_onehot[len(settings.list_filters_combination):]
        df = pd.concat([df[list_training_features], FLT_onehot], axis=1)
        # store feature names
        list_training_features = df.columns.values.tolist()
        hf.create_dataset(
            "features",
            (len(list_training_features),),
            dtype=h5py.special_dtype(vlen=str),
        )
        hf["features"][:] = list_training_features
        logging_utils.print_green(
            "Saved features:", ",".join(list_training_features))

        # Save training features to hdf5
        logging_utils.print_green("Save data features to HDF5")
        arr_feat = df[list_training_features].values
        hf["data"].attrs["n_features"] = len(list_training_features)
        for idx, idx_pair in enumerate(
            tqdm(list_start_end, desc="Filling hdf5", ncols=100)
        ):
            arr = arr_feat[idx_pair[0]: idx_pair[1]]
            hf["data"][idx] = np.ravel(arr)



import os
import h5py
import glob
import numpy as np
import pandas as pd
from tqdm import tqdm
from natsort import natsorted

import torch
import torch.nn as nn

from ..utils import data_utils as du
from ..utils import training_utils as tu
from ..utils import logging_utils as lu


def load_plasticc_test(hdf5_file_name, settings):
    """Load data from HDF5
    """

    with h5py.File(hdf5_file_name, "r") as hf:

        n_features = hf["data"].attrs["n_features"]

        arr_data = hf["data"][:]
        arr_SNID = hf["SNID"][:]
        n_samples = arr_data.shape[0]

        # Create a dummy target
        arr_target = np.zeros(n_samples)

        return tu.fill_data_list(
            np.arange(n_samples),
            arr_data,
            arr_target,
            arr_SNID,
            settings,
            n_features,
            "Loading Test Set",
            test=False,
        )


def get_predictions(settings):

    settings.random_length = False
    settings.random_redshift = False

    if "vanilla" in settings.pytorch_model_name:
        settings.num_inference_samples = 1

    # Load RNN model
    rnn = tu.get_model(settings, len(settings.training_features))

    dump_dir = f"{settings.models_dir}/{settings.pytorch_model_name}"
    model_file = f"{dump_dir}/{settings.pytorch_model_name}.pt"

    rnn_state = torch.load(model_file, map_location=lambda storage, loc: storage)
    rnn.load_state_dict(rnn_state)
    rnn.to(settings.device)
    rnn.eval()

    # Load the data
    list_test_hdf5 = glob.glob(os.path.join(settings.processed_dir, "*test_set_*.h5"))
    # Sort the hdf5 files
    list_test_hdf5 = natsorted(list_test_hdf5)

    #####################################
    # Loop over test databases and predict
    #####################################

    list_df = []

    for hdf5_file_name in tqdm(list_test_hdf5, ncols=100):

        # Load test data
        list_data_test = load_plasticc_test(hdf5_file_name, settings)
        num_samples = len(list_data_test)

        # Split into batches
        batch_size = min(2048, num_samples)
        num_batches = num_samples / batch_size
        list_batches = np.array_split(np.arange(num_samples), num_batches)
        for index, batch_idxs in enumerate(list_batches):

            # Get the object ID (or SNID)
            arr_SNID = (
                np.array([list_data_test[i][2] for i in batch_idxs])
                .astype(int)
                .reshape(-1, 1)
            )

            with torch.no_grad():

                packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
                    list_data_test, batch_idxs, settings
                )

                # Predict
                if "vanilla" in settings.pytorch_model_name:
                    out = rnn.forward(packed)
                    arr_preds = nn.functional.softmax(out, dim=-1).data.cpu().numpy()

                elif settings.mean_field_inference is True:
                    out = rnn.forward(packed, mean_field_inference=True)
                    arr_preds = nn.functional.softmax(out, dim=-1).data.cpu().numpy()

                else:

                    # Unpack the array, reshape it for multiple predictions
                    unpacked_tensor, lengths = nn.utils.rnn.pad_packed_sequence(packed)

                    L, B, D = unpacked_tensor.shape

                    # Copy the data num_inference_sample times for predictions over a new dimension
                    unpacked_tensor = (
                        unpacked_tensor.view(L, B, 1, D)
                        .expand(L, B, settings.num_inference_samples, D)
                        .clone()
                    )
                    lengths = (
                        lengths.view(B, 1)
                        .expand(B, settings.num_inference_samples)
                        .clone()
                    )

                    # Then reflatten the array to pass them to the model
                    unpacked_tensor = unpacked_tensor.view(
                        L, B * settings.num_inference_samples, D
                    )
                    lengths = lengths.view(B * settings.num_inference_samples)

                    # Pack the data
                    packed = nn.utils.rnn.pack_padded_sequence(unpacked_tensor, lengths)

                    out = rnn.forward(packed)

                    # Obtain predictions, reshape them from (B * settings.num_inference_samples , nb_classes)
                    # to (B, settings.num_inference_samples, nb_classes) and take the mean
                    arr_preds = (
                        nn.functional.softmax(out, dim=-1)
                        .data.cpu()
                        .numpy()
                        .reshape((B, settings.num_inference_samples, -1))
                        .mean(axis=1)
                    )

                # Revert sorting
                arr_preds = arr_preds[idxs_rev_sort]

                df_pred = pd.DataFrame(
                    data=arr_preds,
                    columns=[f"class_{i}" for i in range(arr_preds.shape[-1])],
                )
                df_pred["object_id"] = arr_SNID

            list_df.append(df_pred)

    # Concat all predictions
    df = pd.concat(list_df)
    # Change header name
    df = df.rename(
        columns={
            f"class_{value}": f"class_{key}"
            for (key, value) in du.DICT_PLASTICC_CLASS.items()
        }
    )
    df = df.sort_values("object_id")

    # Add class 99 column
    df["class_99"] = np.zeros(len(df))

    # Reorder columns
    id_column = ["object_id"]
    class_columns = [f"class_{key}" for key in du.DICT_PLASTICC_CLASS.keys()]
    df[class_columns] = df[class_columns].astype(np.float16)
    columns = id_column + class_columns
    df = df[columns]

    # Rework class 99
    arr_preds = df[[c for c in class_columns if c != "class_99"]].values

    # max_pred = arr_preds.max(axis=1)
    # arr_99 = df["class_99"].values
    # # Heuristic: when the model is confident in its predictions, leave arr_99 to 0
    # arr_99[max_pred >= 0.8] = 0
    # # Otherwise, assign it to 1 - max_pred
    # arr_99[max_pred < 0.8] = (1 - max_pred)[max_pred < 0.8]
    # df["class_99"] = arr_99

    # Olivier's code
    preds_99 = np.ones(arr_preds.shape[0])
    for i in range(arr_preds.shape[1]):
        preds_99 *= 1 - arr_preds[:, i]

    df["class_99"] = 0.14 * preds_99 / np.mean(preds_99)

    # Renormalize proba
    arr = df[class_columns].values
    arr = arr / arr.sum(1, keepdims=True)
    df[class_columns] = arr

    # Load the metadata test set and check we do have all predictions
    df_meta = pd.read_csv(os.path.join(settings.raw_dir, "test_set_metadata.csv"))

    sorted_id = np.sort(df.object_id.unique())
    sorted_id_from_meta = np.sort(df_meta.object_id.unique())

    # Check we do have all the snid in the prediction set
    assert np.all(sorted_id == sorted_id_from_meta)

    # Save predictions
    prediction_file = f"{dump_dir}/rnn_predictions_{settings.pytorch_model_name}.csv"
    df.to_csv(prediction_file, index=False)



import os
import pytest
from pathlib import Path
from tests.test_utils import call_cmd
from tests.test_utils import testmanager


@pytest.mark.parametrize(
    "dataset, redshift, norm, nb_classes",
    [
        ("photometry", None, "global", 2),
        ("photometry", None, "global", 3),
        ("photometry", None, "global", 7),
        ("photometry", None, "none", 2),
        ("photometry", None, "perfilter", 2),
        ("photometry", None, "perfilter", 2),
        ("photometry", None, "perfilter", 2),
        ("photometry", "zpho", "none", 2),
        ("photometry", "zspe", "none", 2),
        ("saltfit", None, "global", 2),
        ("saltfit", None, "global", 3),
        ("saltfit", None, "global", 7),
        ("saltfit", None, "none", 2),
        ("saltfit", None, "perfilter", 2),
        ("saltfit", None, "perfilter", 2),
        ("saltfit", None, "perfilter", 2),
        ("saltfit", "zpho", "none", 2),
        ("saltfit", "zspe", "none", 2),
    ],
)
@testmanager()
def test_rnn(dataset, redshift, norm, nb_classes):

    cmd = (
        f"python run.py --train_rnn --source_data {dataset} "
        f"--dump_dir tests/dump "
        f"--nb_epoch 1 --norm {norm} --nb_classes {nb_classes}"
    )
    if redshift is not None:
        cmd += f" --redshift {redshift}"
    call_cmd(cmd)

    cmd = (
        f"python run.py --validate_rnn --source_data {dataset} "
        f"--dump_dir tests/dump "
        f"--norm {norm} --nb_classes {nb_classes}"
    )
    if redshift is not None:
        cmd += f" --redshift {redshift}"
    call_cmd(cmd)

    # Check we do have predictions saved
    dir_path = os.path.dirname(os.path.realpath(__file__))
    assert len([e for e in (Path(dir_path) / "dump/models").glob("*/PRED*")]) == 1


if __name__ == "__main__":

    test_rnn("saltfit", None, "perfilter", 2)



import math
import numbers
import warnings

import torch
import torch.nn as nn
import torch.nn.functional as F


# mean_field_inference: instead of sampling a weight as in W = N(mu, sigma)
# we always set W = mu


class BayesianRNN(nn.Module):

    def __init__(self, input_size, settings):
        super(BayesianRNN, self).__init__()

        # Params
        self.layer_type = settings.layer_type
        self.hidden_size = settings.hidden_dim
        self.output_size = settings.nb_classes
        self.num_layers = settings.num_layers
        self.bidirectional = settings.bidirectional
        self.use_cuda = settings.use_cuda
        self.rnn_output_option = settings.rnn_output_option

        self.prior = Prior(settings.pi, settings.log_sigma1, settings.log_sigma2)
        self.prior_output = Prior(
            settings.pi, settings.log_sigma1_output, settings.log_sigma2_output
        )

        bidirectional_factor = 2 if self.bidirectional is True else 1
        last_input_size = (
            self.hidden_size * bidirectional_factor
            if self.rnn_output_option == "mean"
            else self.hidden_size * bidirectional_factor * self.num_layers
        )

        # Layers / nn objects
        self.rnn_layer = BayesRNNBase(
            self.layer_type.upper(),
            input_size,
            self.hidden_size,
            self.prior,
            num_layers=self.num_layers,
            bidirectional=self.bidirectional,
            mu_lower=-0.05,
            mu_upper=0.05,
            rho_lower=math.log(
                math.exp(self.prior.sigma_mix / settings.rho_scale_lower) - 1.0
            ),
            rho_upper=math.log(
                math.exp(self.prior.sigma_mix / settings.rho_scale_upper) - 1.0
            ),
        )

        self.output_layer = BayesLinear(
            last_input_size,
            self.output_size,
            self.prior_output,
            mu_lower=-0.05,
            mu_upper=0.05,
            rho_lower=math.log(
                math.exp(self.prior_output.sigma_mix / settings.rho_scale_lower_output)
                - 1.0
            ),
            rho_upper=math.log(
                math.exp(self.prior_output.sigma_mix / settings.rho_scale_upper_output)
                - 1.0
            ),
        )

    def forward(self, x, mean_field_inference=False):
        # out = packed output from last layer
        # out has dim (seq_len, batch_size, hidden_size) when unpacked
        # hidden = (hn, cn) for lstm
        # hidden = hn for GRU and RNN
        # hn has dim (num_layers * num_directions, batch, hidden_size)
        # cn has dim (num_layers * num_directions, batch, hidden_size)
        # assuming num_directions = 1, num_layers = 2 :
        # hn[-1, -1] == out[len, -1] where len is the len of the seq at batch index == -1

        x, hidden = self.rnn_layer(x, mean_field_inference=mean_field_inference)

        # Output options
        # Standard: all layers, only end of pass
        #    - take last pass in all layers (hidden)
        #    - reshape and apply dropout
        #    - use h20 to obtain output (h2o input: hidden_size*num_layers*bi)
        # Mean: last layer, mean on sequence
        #    - take packed output from last layer (out) that contains all time steps for the last layer
        #    - find where padding was done and create a mask for those values, apply this mask
        #    - take a mean for the whole sequence (time_steps)
        #    - use h2o to obtain output (beware! it is only one layer deep since it is the last one only)

        if self.rnn_output_option == "standard":
            # Special case for lstm where hidden = (h, c)
            if self.layer_type == "lstm":
                hn = hidden[0]
            else:
                hn = hidden

            hn = hn.permute(1, 2, 0).contiguous()
            # hn is (num_layers * num_directions, batch, hidden_size)
            batch_size = hn.shape[0]
            # hn now is (batch, hidden size, num_layers * num_directions)
            x = hn.view(batch_size, -1)
            # x is (batch, hidden size * num_layers * num_directions)

        if self.rnn_output_option == "mean":
            if isinstance(x, torch.nn.utils.rnn.PackedSequence):
                x, lens = torch.nn.utils.rnn.pad_packed_sequence(x)
                # x is (seq_len, batch, hidden size)

                # take mean over seq_len
                x = x.sum(0) / lens.unsqueeze(-1).float().to(x.device)
                # x is (batch, hidden_size)
            else:
                x = x.mean(0)
                # x is (batch, hidden_size)

        # Final projection layer
        output = self.output_layer(x, mean_field_inference=mean_field_inference)

        # Compute KL
        self.kl = self.rnn_layer.kl + self.output_layer.kl

        return output


class BayesRNNBase(nn.Module):

    def __init__(
        self,
        mode,
        input_size,
        hidden_size,
        prior,
        num_layers=1,
        bias=True,
        batch_first=False,
        dropout=0,
        bidirectional=False,
        mu_lower=-0.05,
        mu_upper=0.05,
        rho_lower=-1,
        rho_upper=-4,
    ):
        super(BayesRNNBase, self).__init__()
        self.mode = mode
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.prior = prior
        self.num_layers = num_layers
        self.bias = bias
        self.batch_first = batch_first
        self.dropout = dropout
        self.dropout_state = {}
        self.bidirectional = bidirectional
        num_directions = 2 if bidirectional else 1

        if (
            not isinstance(dropout, numbers.Number)
            or not 0 <= dropout <= 1
            or isinstance(dropout, bool)
        ):
            raise ValueError(
                "dropout should be a number in range [0, 1] "
                "representing the probability of an element being "
                "zeroed"
            )
        if dropout > 0 and num_layers == 1:
            warnings.warn(
                "dropout option adds dropout after all but last "
                "recurrent layer, so non-zero dropout expects "
                "num_layers greater than 1, but got dropout={} and "
                "num_layers={}".format(dropout, num_layers)
            )

        if mode == "LSTM":
            gate_size = 4 * hidden_size
        elif mode == "GRU":
            gate_size = 3 * hidden_size
        else:
            gate_size = hidden_size

        # self._all_weights = []
        for layer in range(num_layers):
            for direction in range(num_directions):
                layer_input_size = (
                    input_size if layer == 0 else hidden_size * num_directions
                )

                w_ih_mu, w_ih_rho = get_bbb_variable(
                    (gate_size, layer_input_size),
                    mu_lower,
                    mu_upper,
                    rho_lower,
                    rho_upper,
                )
                w_hh_mu, w_hh_rho = get_bbb_variable(
                    (gate_size, hidden_size), mu_lower, mu_upper, rho_lower, rho_upper
                )

                b_ih_mu, b_ih_rho = get_bbb_variable(
                    (gate_size,), mu_lower, mu_upper, rho_lower, rho_upper
                )
                b_hh_mu, b_hh_rho = get_bbb_variable(
                    (gate_size,), mu_lower, mu_upper, rho_lower, rho_upper
                )

                layer_params = (
                    w_ih_mu,
                    w_ih_rho,
                    w_hh_mu,
                    w_hh_rho,
                    b_ih_mu,
                    b_ih_rho,
                    b_hh_mu,
                    b_hh_rho,
                )

                suffix = "_reverse" if direction == 1 else ""
                param_names = [
                    "weight_ih_mu_l{}{}",
                    "weight_ih_rho_l{}{}",
                    "weight_hh_mu_l{}{}",
                    "weight_hh_rho_l{}{}",
                ]

                if bias:
                    param_names += [
                        "bias_ih_mu_l{}{}",
                        "bias_ih_rho_l{}{}",
                        "bias_hh_mu_l{}{}",
                        "bias_hh_rho_l{}{}",
                    ]
                param_names = [x.format(layer, suffix) for x in param_names]

                for name, param in zip(param_names, layer_params):
                    setattr(self, name, param)
                # self._all_weights.append(param_names)

        self.flatten_parameters()
        self.kl = None

    def flatten_parameters(self):
        """Resets parameter data pointer so that they can use faster code paths.
        """
        self._data_ptrs = []
        self._param_buf_size = 0
        return

    def _apply(self, fn):
        ret = super(BayesRNNBase, self)._apply(fn)
        self.flatten_parameters()
        return ret

    def check_forward_args(self, input, hidden, batch_sizes):
        is_input_packed = batch_sizes is not None
        expected_input_dim = 2 if is_input_packed else 3
        if input.dim() != expected_input_dim:
            raise RuntimeError(
                "input must have {} dimensions, got {}".format(
                    expected_input_dim, input.dim()
                )
            )
        if self.input_size != input.size(-1):
            raise RuntimeError(
                "input.size(-1) must be equal to input_size. Expected {}, got {}".format(
                    self.input_size, input.size(-1)
                )
            )

        if is_input_packed:
            mini_batch = int(batch_sizes[0])
        else:
            mini_batch = input.size(0) if self.batch_first else input.size(1)

        num_directions = 2 if self.bidirectional else 1
        expected_hidden_size = (
            self.num_layers * num_directions,
            mini_batch,
            self.hidden_size,
        )

        def check_hidden_size(
            hx, expected_hidden_size, msg="Expected hidden size {}, got {}"
        ):
            if tuple(hx.size()) != expected_hidden_size:
                raise RuntimeError(msg.format(expected_hidden_size, tuple(hx.size())))

        if self.mode == "LSTM":
            check_hidden_size(
                hidden[0], expected_hidden_size, "Expected hidden[0] size {}, got {}"
            )
            check_hidden_size(
                hidden[1], expected_hidden_size, "Expected hidden[1] size {}, got {}"
            )
        else:
            check_hidden_size(hidden, expected_hidden_size)

    def forward(self, input, hx=None, mean_field_inference=False):

        is_packed = isinstance(input, nn.utils.rnn.PackedSequence)
        if is_packed:
            input, batch_sizes = input
            max_batch_size = int(batch_sizes[0])
        else:
            batch_sizes = None
            max_batch_size = input.size(0) if self.batch_first else input.size(1)

        if hx is None:
            num_directions = 2 if self.bidirectional else 1
            hx = input.new_zeros(
                self.num_layers * num_directions,
                max_batch_size,
                self.hidden_size,
                requires_grad=False,
            )
            if self.mode == "LSTM":
                hx = (hx, hx)

        has_flat_weights = (
            list(p.data.data_ptr() for p in self.parameters()) == self._data_ptrs
        )
        if has_flat_weights:
            first_data = next(self.parameters()).data
            assert first_data.storage().size() == self._param_buf_size
            flat_weight = first_data.new().set_(
                first_data.storage(), 0, torch.Size([self._param_buf_size])
            )
        else:
            flat_weight = None

        self.check_forward_args(input, hx, batch_sizes)
        func = self._backend.RNN(
            self.mode,
            self.input_size,
            self.hidden_size,
            num_layers=self.num_layers,
            batch_first=self.batch_first,
            dropout=self.dropout,
            train=self.training,
            bidirectional=self.bidirectional,
            dropout_state=self.dropout_state,
            variable_length=is_packed,
            flat_weight=flat_weight,
        )

        # Format weights for BBB
        all_weights = []
        num_layers = self.num_layers
        num_directions = 2 if self.bidirectional else 1
        self.kl = 0
        # Loop over layers
        for layer_idx in range(num_layers):
            # Loop over directions
            for direction in range(num_directions):

                suffix = "_reverse" if direction == 1 else ""

                w_ih_mean = getattr(self, f"weight_ih_mu_l{layer_idx}{suffix}")
                w_ih_rho = getattr(self, f"weight_ih_rho_l{layer_idx}{suffix}")
                w_ih_sigma = torch.nn.functional.softplus(w_ih_rho) + 1e-5

                w_hh_mean = getattr(self, f"weight_hh_mu_l{layer_idx}{suffix}")
                w_hh_rho = getattr(self, f"weight_hh_rho_l{layer_idx}{suffix}")
                w_hh_sigma = torch.nn.functional.softplus(w_hh_rho) + 1e-5

                b_ih_mean = getattr(self, f"bias_ih_mu_l{layer_idx}{suffix}")
                b_ih_rho = getattr(self, f"bias_ih_rho_l{layer_idx}{suffix}")
                b_ih_sigma = torch.nn.functional.softplus(b_ih_rho) + 1e-5

                b_hh_mean = getattr(self, f"bias_hh_mu_l{layer_idx}{suffix}")
                b_hh_rho = getattr(self, f"bias_hh_rho_l{layer_idx}{suffix}")
                b_hh_sigma = torch.nn.functional.softplus(b_hh_rho) + 1e-5

                if mean_field_inference:
                    weight_ih = w_ih_mean
                    weight_hh = w_hh_mean
                    if self.bias:
                        bias_ih = b_ih_mean
                        bias_hh = b_hh_mean
                else:

                    # Sample weights from normal distribution
                    eps_ih = w_ih_mean.data.new(w_ih_mean.size()).normal_(0.0, 1.0)
                    weight_ih = w_ih_mean + eps_ih * w_ih_sigma

                    eps_hh = w_hh_mean.data.new(w_hh_mean.size()).normal_(0.0, 1.0)
                    weight_hh = w_hh_mean + eps_hh * w_hh_sigma

                # Compute KL divergence
                self.kl += compute_KL(weight_ih, w_ih_mean, w_ih_sigma, self.prior)
                self.kl += compute_KL(weight_hh, w_hh_mean, w_hh_sigma, self.prior)

                weights = [weight_ih, weight_hh]

                # Get biases
                if self.bias:

                    eps_ih = b_ih_mean.data.new(b_ih_mean.size()).normal_(0.0, 1.0)
                    bias_ih = b_ih_mean + eps_ih * b_ih_sigma

                    eps_hh = b_hh_mean.data.new(b_hh_mean.size()).normal_(0.0, 1.0)
                    bias_hh = b_hh_mean + eps_hh * b_hh_sigma

                    self.kl += compute_KL(bias_ih, b_ih_mean, b_ih_sigma, self.prior)
                    self.kl += compute_KL(bias_hh, b_hh_mean, b_hh_sigma, self.prior)

                    weights += [bias_ih, bias_hh]

                all_weights.append(weights)

        output, hidden = func(input, all_weights, hx, batch_sizes)
        if is_packed:
            output = nn.utils.rnn.PackedSequence(output, batch_sizes)

        return output, hidden

    def extra_repr(self):
        s = "{input_size}, {hidden_size}"
        if self.num_layers != 1:
            s += ", num_layers={num_layers}"
        if self.bias is not True:
            s += ", bias={bias}"
        if self.batch_first is not False:
            s += ", batch_first={batch_first}"
        if self.dropout != 0:
            s += ", dropout={dropout}"
        if self.bidirectional is not False:
            s += ", bidirectional={bidirectional}"
        return s.format(**self.__dict__)


class BayesLSTM(BayesRNNBase):

    def __init__(self, *args, **kwargs):
        super(BayesLSTM, self).__init__("LSTM", *args, **kwargs)


class BayesGRU(BayesRNNBase):

    def __init__(self, *args, **kwargs):
        super(BayesGRU, self).__init__("GRU", *args, **kwargs)


class BayesLinear(nn.Module):

    def __init__(
        self, in_features, out_features, prior, mu_lower, mu_upper, rho_lower, rho_upper
    ):
        super(BayesLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.prior = prior

        mu, rho = get_bbb_variable(
            (out_features, in_features), mu_lower, mu_upper, rho_lower, rho_upper
        )

        bias = nn.Parameter(torch.Tensor(out_features))
        bias.data.fill_(0.0)

        self.mu = mu
        self.rho = rho
        self.bias = bias
        self.kl = None

    def forward(self, input, mean_field_inference=False):

        # Sample weight
        mean = self.mu
        sigma = F.softplus(self.rho) + 1e-5

        if mean_field_inference:
            weights = mean
        else:
            # Sample weights from normal distribution
            # This way of creating the epsilon variable is faster than
            # from numpy or torch.randn or FloatTensor.normal_ when mean is already
            # on the GPU
            eps = mean.data.new(mean.size()).normal_(0.0, 1.0)
            weights = mean + eps * sigma

        logits = F.linear(input, weights, self.bias)

        # Compute KL divergence
        self.kl = compute_KL(weights, mean, sigma, self.prior)

        return logits

    def __repr__(self):
        return (
            self.__class__.__name__
            + " ("
            + str(self.in_features)
            + " -> "
            + str(self.out_features)
            + ")"
        )


class BayesBiasLinear(nn.Module):

    def __init__(
        self, in_features, out_features, prior, mu_lower, mu_upper, rho_lower, rho_upper
    ):
        super(BayesBiasLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.prior = prior

        self.W_mu, self.W_rho = get_bbb_variable(
            (out_features, in_features), mu_lower, mu_upper, rho_lower, rho_upper
        )

        self.b_mu, self.b_rho = get_bbb_variable(
            (out_features,), mu_lower, mu_upper, rho_lower, rho_upper
        )

    def forward(self, X, mean_field_inference=False):
        # Sample weight
        W_mean = self.W_mu
        W_sigma = F.softplus(self.W_rho) + 1e-5

        b_mean = self.b_mu
        b_sigma = F.softplus(self.b_rho) + 1e-5

        if mean_field_inference:
            weights = W_mean
            biases = b_mean
        else:
            # Sample weights from normal distribution
            # This way of creating the epsilon variable is faster than
            # from numpy or torch.randn or FloatTensor.normal_ when mean is already
            # on the GPU
            W_eps = W_mean.data.new(W_mean.size()).normal_(0.0, 1.0)
            weights = W_mean + W_eps * W_sigma

            b_eps = b_mean.data.new(b_mean.size()).normal_(0.0, 1.0)
            biases = b_mean + b_eps * b_sigma

        logits = F.linear(X, weights, biases)

        # Compute KL divergence
        self.kl = compute_KL(weights, W_mean, W_sigma, self.prior)
        self.kl += compute_KL(biases, b_mean, b_sigma, self.prior)

        return logits


class BayesEmbedding(nn.Module):

    def __init__(
        self,
        num_embeddings,
        embedding_dim,
        prior,
        mu_lower,
        mu_upper,
        rho_lower,
        rho_upper,
    ):
        super(BayesEmbedding, self).__init__()

        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim
        self.prior = prior
        self.max_norm = None
        self.norm_type = 2
        self.scale_grad_by_freq = False
        self.sparse = False
        self.padding_idx = -1

        mu, rho = get_bbb_variable(
            [num_embeddings, embedding_dim], mu_lower, mu_upper, rho_lower, rho_upper
        )

        self.mu = mu
        self.rho = rho
        self.kl = None

    def forward(self, input, mean_field_inference=False):

        # Sample weight
        mean = self.mu
        sigma = F.softplus(self.rho) + 1E-5

        if mean_field_inference:
            weights = mean
        else:
            # This way of creating the epsilon variable is faster than
            # from numpy or torch.randn or FloatTensor.normal_ when mean is already
            # on the GPU
            eps = mean.data.new(mean.size()).normal_(0., 1.)
            weights = mean + eps * sigma

        # Compute KL divergence
        self.kl = compute_KL(weights, mean, sigma, self.prior)

        after_embed = F.embedding(
            input,
            weights,
            self.padding_idx,
            self.max_norm,
            self.norm_type,
            self.scale_grad_by_freq,
            self.sparse,
        )

        return after_embed

    def __repr__(self):
        s = "{name}({num_embeddings}, {embedding_dim}"
        if self.padding_idx is not None:
            s += ", padding_idx={padding_idx}"
        if self.max_norm is not None:
            s += ", max_norm={max_norm}"
        if self.norm_type != 2:
            s += ", norm_type={norm_type}"
        if self.scale_grad_by_freq is not False:
            s += ", scale_grad_by_freq={scale_grad_by_freq}"
        if self.sparse is not False:
            s += ", sparse=True"
        s += ")"
        return s.format(name=self.__class__.__name__, **self.__dict__)


class Prior(object):

    def __init__(self, pi=0.25, log_sigma1=-1.0, log_sigma2=-7.0):
        self.pi_mixture = pi
        self.log_sigma1 = log_sigma1
        self.log_sigma2 = log_sigma2
        self.sigma1 = math.exp(log_sigma1)
        self.sigma2 = math.exp(log_sigma2)

        self.sigma_mix = math.sqrt(
            pi * math.pow(self.sigma1, 2) + (1.0 - pi) * math.pow(self.sigma2, 2)
        )


def get_bbb_variable(shape, mu_lower, mu_upper, rho_lower, rho_upper):

    mu = nn.Parameter(torch.Tensor(*shape))
    rho = nn.Parameter(torch.Tensor(*shape))

    # Initialize
    mu.data.uniform_(mu_lower, mu_upper)
    rho.data.uniform_(rho_lower, rho_upper)

    return mu, rho


def logsumexp(x, dim):
    """Logsumexp trick to avoid overflow in a log of sum of exponential expression

    Args:
        x (Tensor): the input on which to compute the log of sum of exponential

    Returns:
        logsum (Tensor): the computed log of sum of exponential
    """

    assert x.dim() == 2
    x_max, x_max_idx = x.max(dim=dim, keepdim=True)
    logsum = x_max + torch.log((x - x_max).exp().sum(dim=dim, keepdim=True))
    return logsum


def compute_KL(x, mu, sigma, prior):

    posterior = torch.distributions.Normal(mu.view(-1), sigma.view(-1))
    log_posterior = posterior.log_prob(x.view(-1)).sum()

    if x.is_cuda:
        n1 = torch.distributions.Normal(
            torch.tensor([0.0]).cuda(), torch.tensor([prior.sigma1]).cuda()
        )
        n2 = torch.distributions.Normal(
            torch.tensor([0.0]).cuda(), torch.tensor([prior.sigma2]).cuda()
        )
    else:
        n1 = torch.distributions.Normal(0.0, prior.sigma1)
        n2 = torch.distributions.Normal(0.0, prior.sigma2)

    mix1 = torch.sum(n1.log_prob(x)) + math.log(prior.pi_mixture)
    mix2 = torch.sum(n2.log_prob(x)) + math.log(1.0 - prior.pi_mixture)
    prior_mix = torch.stack([mix1, mix2])
    log_prior = torch.logsumexp(prior_mix, 0)

    return log_posterior - log_prior



import numpy as np
from pathlib import Path
from ..utils import data_utils
from ..utils import training_utils
from ..utils import logging_utils


def get_predictions(settings, model_file=None):
    """Test random forest models on independent test set

    Features are stored in a .FITRES file found in data_dir
    Use predefined splits to select test set
    Save predicted target and probabilities to preds_dir

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters
        model_file (str): path to saved randomforest model
    """

    assert settings.source_data == "saltfit", logging_utils.str_to_redstr(
        "Only salfit is a valid data source for random forest"
    )
    assert settings.nb_classes == 2, logging_utils.str_to_redstr(
        "Binary classification is the only task allowed for RandomForest"
    )

    if model_file is None:
        dump_dir = f"{settings.models_dir}/{settings.randomforest_model_name}"
        model_file = f"{dump_dir}/{settings.randomforest_model_name}.pickle"
    else:
        dump_dir = Path(model_file).parent

    if settings.override_source_data is not None:
        settings.source_data = settings.override_source_data
        settings.set_pytorch_model_name()

    prediction_file = (
        f"{dump_dir}/PRED_{settings.randomforest_model_name}.pickle"
    )

    # load data
    df_test = data_utils.load_fitfile(settings)

    # Load the dataframe containing  the list of the independent test SNID
    sn_df = data_utils.load_HDF5_SNinfo(settings)
    df_SNID = sn_df[sn_df["dataset_saltfit_2classes"] == 2][
        ["SNID", "PEAKMJD", "PEAKMJDNORM", "SIM_REDSHIFT_CMB", "SNTYPE"]
    ]

    # Make sure SNID is of type int
    df_SNID["SNID"] = df_SNID["SNID"].astype(int)
    df_test["SNID"] = df_test["SNID"].astype(int)

    # Select only IDs that are in df_SNID and df_test
    df_test = df_test.merge(df_SNID, on="SNID")

    # Add redshift features if required by settings
    df_test = data_utils.add_redshift_features(settings, df_test)

    # Load randomforest model
    clf = training_utils.load_randomforest_model(settings, model_file=model_file)
    # Make predictions and save to dataframe
    X = df_test[settings.randomforest_features].values
    y_pred_proba = clf.predict_proba(X)
    y_pred_class = np.argmax(y_pred_proba, axis=1)

    df_test["all_class0"] = y_pred_proba[:, 0]
    df_test["all_class1"] = y_pred_proba[:, 1]
    df_test["predicted_target"] = y_pred_class
    df_test["target"] = df_test["target_2classes"]

    # Cache results for future re-use
    list_features_save = [
        "SNID",
        "all_class0",
        "all_class1",
        "predicted_target",
        "target",
    ]
    df_test[list_features_save].to_pickle(prediction_file)

    return prediction_file



import os
import pytest
from pathlib import Path
from tests.test_utils import call_cmd
from tests.test_utils import testmanager


@pytest.mark.parametrize("redshift", [None, "zpho", "zspe"])
@testmanager()
def test_rf(redshift):

    cmd = (
        "python run.py --train_rf --max_depth 1 --max_features 1 --n_estimators 1 --dump_dir tests/dump"
    )
    if redshift is not None:
        cmd += f" --redshift {redshift}"
    call_cmd(cmd)
    cmd = "python run.py --validate_rf --dump_dir tests/dump"
    if redshift is not None:
        cmd += f" --redshift {redshift}"
    call_cmd(cmd)

    # Check we do have predictions saved
    dir_path = os.path.dirname(os.path.realpath(__file__))
    assert (
        len([e for e in (Path(dir_path) / "dump/models").glob("*forest*/PRED*")]) == 1
    )


if __name__ == "__main__":

    test_rf(None)



import os
import h5py
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from collections import defaultdict

# plt.switch_backend("agg")
import matplotlib.gridspec as gridspec
from ..utils.data_utils import DICT_PLASTICC_FILTERS, PLASTICC_FILTERS
from ..data import make_dataset_plasticc


def plot_histogram(df, settings):

    list_fields = ["delta_time"]
    list_fields += ["FLUXCAL_%s" % k for k in ["r", "g", "z", "i"]]
    list_fields += ["FLUXCALERR_%s" % k for k in ["r", "g", "z", "i"]]

    plt.figure(figsize=(20, 20))
    gs = gridspec.GridSpec(3, 3, hspace=0.4)
    for i in range(9):
        ax = plt.subplot(gs[i])
        arr = df[list_fields[i]].values
        ax.hist(arr, bins=500)
        # ax.set_xlim([minx, maxx])
        ax.set_title(list_fields[i], fontsize=22)
        ax.set_yscale("log")
    plt.savefig(os.path.join(settings.fig_dir, "histogram_feature_distribution.png"))


def plot_lightcurves(df, SNIDs):
    """Utility for gridspec of lightcruves

    Args:
        df (pandas.DataFrame): dataframe holding the data
        SNIDs (np.array or list): holds lightcurve ids
    """

    plt.figure(figsize=(20, 10))
    plt.suptitle("Sample of SN Ia light curves")
    gs = gridspec.GridSpec(4, 4, hspace=0.4)

    for i in range(16):
        ax = plt.subplot(gs[i])

        SNID = SNIDs[i]
        df_temp = df.loc[SNID]

        # Prepare plotting data in a dict
        d = {flt: {"FLUXCAL": [], "FLUXCALERR": [], "MJD": []} for flt in FILTERS}

        current_time = 0
        for idx in range(len(df_temp)):
            flt = df_temp.FLT.values[idx]
            d[flt]["FLUXCAL"].append(df_temp.FLUXCAL.values[idx])
            d[flt]["FLUXCALERR"].append(df_temp.FLUXCALERR.values[idx])
            current_time += df_temp.delta_time.values[idx]
            d[flt]["MJD"].append(current_time)

        for flt in d.keys():
            time = d[flt]["MJD"]
            # Only plot a time series if it's non empty
            if len(time) > 0:
                flux = d[flt]["FLUXCAL"]
                fluxerr = d[flt]["FLUXCALERR"]
                ax.errorbar(time, flux, yerr=fluxerr, label=f"Filter {flt}")

        ax.set_title(SNID, fontsize=18)
        ax.legend(loc="best")
        ax.set_aspect("auto")

    plt.savefig("pickle.svg")


def plot_lightcurves_from_hdf5(settings, SNID_idxs):

    with h5py.File(settings.hdf5_file_name, "r") as hf:

        features = hf["features"][:].astype(str)
        n_features = len(features)

        plt.figure(figsize=(20, 10))
        gs = gridspec.GridSpec(4, 4, hspace=0.4)

        for idx, SNID_idx in enumerate(SNID_idxs):

            ax = plt.subplot(gs[idx])

            SNID = hf["SNID"][SNID_idx]
            PEAKMJD = str(hf["PEAKMJD"][SNID_idx])
            PEAKMJDNORM = hf["PEAKMJDNORM"][SNID_idx]
            data = hf["data"][SNID_idx].reshape(-1, n_features)

            df = pd.DataFrame(data, columns=features)

            non_filter_columns = [
                "FLUXCAL_g",
                "FLUXCAL_i",
                "FLUXCAL_r",
                "FLUXCAL_z",
                "FLUXCALERR_g",
                "FLUXCALERR_i",
                "FLUXCALERR_r",
                "FLUXCALERR_z",
                "delta_time",
                "HOSTGAL_PHOTOZ",
                "HOSTGAL_PHOTOZ_ERR",
                "HOSTGAL_SPECZ",
                "HOSTGAL_SPECZ_ERR",
            ]

            filter_columns = [
                c for c in df.columns.values if c not in non_filter_columns
            ]

            present_filters = df[filter_columns].transpose().idxmax().values
            list_present_filters = [set(f) for f in present_filters]

            max_y = -float("Inf")
            min_y = float("Inf")

            for FLT in FILTERS:
                idxs = np.array(
                    [i for i in range(len(df)) if FLT in list_present_filters[i]]
                )
                if len(idxs) == 0:
                    continue
                arr_flux = df[f"FLUXCAL_{FLT}"].values[idxs]
                arr_fluxerr = df[f"FLUXCALERR_{FLT}"].values[idxs]
                arr_time = df["delta_time"].cumsum().values[idxs]
                ax.errorbar(arr_time, arr_flux, yerr=arr_fluxerr, label=f"Filter {FLT}")

                if np.max(arr_flux) > max_y:
                    max_y = np.max(arr_flux)

                if np.min(arr_flux) < min_y:
                    min_y = np.min(arr_flux)

            ax.plot(
                [PEAKMJDNORM, PEAKMJDNORM], [min_y, max_y], color="k", linestyle="--"
            )
            ax.set_title(f"{SNID} -- {PEAKMJD}", fontsize=18)
            ax.legend(loc="best")
            ax.set_aspect("auto")

        plt.savefig("figure.svg")


def visualize_plasticc(settings):

    # Check the data has been created
    settings.check_data_exists()

    d_rev = {value: key for (key, value) in DICT_PLASTICC_FILTERS.items()}

    # Set a random seed
    np.random.seed()

    with h5py.File(settings.hdf5_file_name, "r") as hf:
        SNID_idxs = np.random.permutation(hf["SNID"].shape[0])[:16]
        SNIDs = hf["SNID"][:][SNID_idxs]

    df = pd.read_csv(os.path.join(settings.raw_dir, "training_set.csv"))
    df_meta = pd.read_csv(os.path.join(settings.raw_dir, "training_set_metadata.csv"))
    df = df.merge(df_meta, on="object_id", how="left")

    df_preproc = make_dataset_plasticc.preprocess_plasticc(settings, df.copy())
    df_pivot = make_dataset_plasticc.pivot_dataframe(df_preproc.copy()).reset_index()

    for fig_idx, (SNID, SNID_idx) in enumerate(zip(SNIDs, SNID_idxs)):

        fig = plt.figure()
        gs = gridspec.GridSpec(4, 1)

        #########################
        # BEFORE ANY PROCESSING
        #########################

        ax = plt.subplot(gs[0])

        # Loop over data and plot
        df_tmp = df[df.object_id == SNID]
        d_flux = defaultdict(list)
        first_time = df_tmp.mjd.values[0]
        unique_flts = set([DICT_PLASTICC_FILTERS[f] for f in df_tmp.passband.unique()])
        for i in range(df_tmp.shape[0]):
            flux, fluxerr, flt, time = df_tmp[
                ["flux", "flux_err", "passband", "mjd"]
            ].values[i, :]
            flt = DICT_PLASTICC_FILTERS[flt]
            d_flux[f"FLUXCAL_{flt}"].append(flux)
            d_flux[f"FLUXCALERR_{flt}"].append(fluxerr)
            d_flux[f"time_{flt}"].append(time - first_time)

        for i, flt in enumerate(sorted(unique_flts)):
            ax.errorbar(
                d_flux[f"time_{flt}"],
                d_flux[f"FLUXCAL_{flt}"],
                d_flux[f"FLUXCALERR_{flt}"],
                label=flt,
                color=f"C{d_rev[flt]}",
            )
        # ax.legend(loc="best")

        #########################
        # After preprocessing
        #########################

        ax = plt.subplot(gs[1])
        # Loop over data and plot
        df_tmp = df_preproc[df_preproc.SNID == SNID]
        d_flux = defaultdict(list)
        first_time = df_tmp.MJD.values[0]
        unique_flts = df_tmp.FLT.unique()
        ax = plt.subplot(gs[1])
        for i in range(df_tmp.shape[0]):
            flux, fluxerr, flt, time = df_tmp[
                ["FLUXCAL", "FLUXCALERR", "FLT", "MJD"]
            ].values[i, :]
            d_flux[f"FLUXCAL_{flt}"].append(flux)
            d_flux[f"FLUXCALERR_{flt}"].append(fluxerr)
            d_flux[f"time_{flt}"].append(time - first_time)

        for i, flt in enumerate(sorted(unique_flts)):
            ax.errorbar(
                d_flux[f"time_{flt}"],
                d_flux[f"FLUXCAL_{flt}"],
                d_flux[f"FLUXCALERR_{flt}"],
                label=flt,
                color=f"C{d_rev[flt]}",
            )
        # ax.legend(loc="best")

        ################################
        # After preprocessing and pivot
        #################################

        ax = plt.subplot(gs[2])
        # Loop over data and plot
        df_tmp_pivot = df_pivot[df_pivot.SNID == SNID]

        df_tmp_pivot["time"] = df_tmp_pivot["delta_time"].cumsum()

        for i, FLT in enumerate(sorted(PLASTICC_FILTERS)):
            arr_flux = df_tmp_pivot[df_tmp_pivot[f"FLUXCAL_{FLT}"] != 0][
                f"FLUXCAL_{FLT}"
            ].values
            arr_fluxerr = df_tmp_pivot[df_tmp_pivot[f"FLUXCAL_{FLT}"] != 0][
                f"FLUXCALERR_{FLT}"
            ].values
            arr_time = df_tmp_pivot[df_tmp_pivot[f"FLUXCAL_{FLT}"] != 0]["time"].values
            if len(arr_flux) == 0:
                continue
            ax.errorbar(
                arr_time,
                arr_flux,
                yerr=arr_fluxerr,
                label=f"Filter {FLT}",
                color=f"C{d_rev[FLT]}",
            )

        ax.set_aspect("auto")

        #########################
        # After HDF5
        #########################

        with h5py.File(settings.hdf5_file_name, "r") as hf:

            features = hf["features"][:].astype(str)
            n_features = len(features)

            ax = plt.subplot(gs[3])

            SNID = hf["SNID"][SNID_idx]
            data = hf["data"][SNID_idx].reshape(-1, n_features)

            df_tmp = pd.DataFrame(data, columns=features)
            df_tmp["time"] = df_tmp.delta_time.cumsum()

            for i, FLT in enumerate(PLASTICC_FILTERS):
                arr_flux = df_tmp[df_tmp[f"FLUXCAL_{FLT}"] != 0][
                    f"FLUXCAL_{FLT}"
                ].values
                arr_fluxerr = df_tmp[df_tmp[f"FLUXCAL_{FLT}"] != 0][
                    f"FLUXCALERR_{FLT}"
                ].values
                arr_time = df_tmp[df_tmp[f"FLUXCAL_{FLT}"] != 0]["time"].values
                ax.errorbar(
                    arr_time,
                    arr_flux,
                    yerr=arr_fluxerr,
                    label=f"Filter {FLT}",
                    color=f"C{d_rev[FLT]}",
                )

            ax.set_title(f"{SNID}", fontsize=18)
            ax.legend(loc="best")
            ax.set_aspect("auto")

        plt.savefig(f"truc_fig_{fig_idx}.svg")
        plt.clf()
        plt.close()
